import copy
import random
from enum import Enum
from typing import List, Tuple

import numpy as np

from experiments.e1.src.evaluation import (
    genotype_to_phenotype,
    TransformerType,
    evaluate_transformer,
    EvaluationConfig,
)
from experiments.e1.src.genotype import Genotype


class SelectionMode(Enum):
    BEST_FITNESS = 0
    WORST_FITNESS = 1


def tournament_selection(
    population: List[Genotype],
    tournament_size: int,
    selection_mode: SelectionMode,
    transformer_type: TransformerType,
    config: EvaluationConfig,
    train_from_scratch: bool,
) -> Tuple[int, Genotype]:
    tournament_genotypes: List[(int, Genotype)] = random.sample(
        list(enumerate(population)), tournament_size
    )

    fitnesses = []
    for genotype in tournament_genotypes:
        if genotype[1].is_dirty() or train_from_scratch:
            phenotype = genotype_to_phenotype(
                genotype[1], transformer_type, config.input_dim, config.output_dim
            )
            train_loss_list, eval_loss_list = evaluate_transformer(
                phenotype, transformer_type, config
            )
            genotype[1].set_loss(
                train_loss_list[-1], eval_loss_list[-1]
            )  # Cache the fitness

        fitness = genotype[1].fitness()
        if fitness is None:
            raise ValueError("Fitness should never be None here")

        fitnesses.append(fitness)

    if selection_mode == SelectionMode.BEST_FITNESS:
        individual_index = np.argmax(fitnesses)
    elif selection_mode == SelectionMode.WORST_FITNESS:
        individual_index = np.argmin(fitnesses)
    else:
        raise ValueError("Unknown genotype selection mode")

    selected_genotype = copy.deepcopy(
        tournament_genotypes[individual_index]
    )  # Important

    return selected_genotype
