import datetime
from enum import Enum

import numpy as np
import pandas as pd
import torch
from torch import nn

from experiments.e1.src.evaluation import (
    EvaluationConfig,
    evaluate_transformer,
    genotype_to_phenotype,
    TransformerType,
)
from experiments.e1.src.genotype import Genotype
from experiments.e1.src.visualization import make_prediction, plot_prediction
from src.components.attention.Attention import Attention
from src.dataset.LibraDataset import LibraCategory
from src.evolutionary.dataset.bootstrap import (
    boostrap_synthetic_dataset,
    boostrap_libra_dataset,
)
from src.training.evaluation import rmse, mae
from src.utils.helpers import mean


class DatasetType(str, Enum):
    SYNTHETIC = "synthetic"
    LIBRA = "libra"


def evaluate(
    dataset_type: DatasetType = DatasetType.SYNTHETIC,
    forecasting_horizon=1,
    transformer_type=TransformerType.ENCODER_DECODER,
    attention_type: Attention = Attention.Passthrough,
    parameters: list = [64, 1, 4, 8, 7, 0.03],
    epochs: int = 5,
    evaluated_examples: int = 1000,
):
    # Datasets
    batch_size = 16
    input_dim = 400
    # NOTE: Set forecasting horizon!
    output_dim = 400 if forecasting_horizon is None else forecasting_horizon
    split_size = 100
    train_test_ratio = 0.8

    synthetic_length = 5000

    libra_categories = [
        LibraCategory.ECONOMICS,
        LibraCategory.FINANCE,
        LibraCategory.HUMAN,
        LibraCategory.NATURE,
    ]
    libra_dataset_filter = [
        [71, 86, 69, 75, 70, 83, 80, 74, 81, 77],
        [63, 96, 49, 39, 71, 18, 99, 66, 82, 17],
        [86, 56, 46, 21, 67, 5, 2, 92, 68, 97],
        [0, 91, 39, 70, 23, 42, 73, 1, 53, 24],
    ]
    libra_min_training_examples = 100
    libra_max_values_per_dataset = 1000

    if dataset_type == DatasetType.SYNTHETIC:
        train_set, test_set = boostrap_synthetic_dataset(
            input_dim,
            output_dim,
            synthetic_length,
            train_test_ratio,
        )
    elif dataset_type == DatasetType.LIBRA:
        train_set, test_set = boostrap_libra_dataset(
            input_dim,
            output_dim,
            libra_categories,
            libra_dataset_filter,
            libra_min_training_examples,
            libra_max_values_per_dataset,
            train_test_ratio,
        )
    else:
        raise ValueError("Unknown dataset type")

    print(f"# Dataset length: {len(train_set) + len(test_set)}")

    # Training
    criterion = nn.SmoothL1Loss(reduction="sum")  # RMSELoss(reduction="sum")
    optimizer = torch.optim.AdamW(
        [torch.zeros(1, 1)], lr=0.001, betas=(0.9, 0.98), eps=1e-9
    )
    device = "cuda" if torch.cuda.is_available() else "cpu"
    evaluation_config = EvaluationConfig(
        batch_size,
        input_dim,
        output_dim,
        epochs,
        train_set,
        test_set,
        criterion,
        optimizer,
        device,
        split_size,
    )

    # Genotype and model
    # NOTE: Fill in the best genotype values
    embedding_dim = parameters[0]
    patch_size = parameters[1]
    head_count = parameters[2]
    layer_count = parameters[3]
    forward_expansion = parameters[4]
    dropout = parameters[5]

    genotype = Genotype.create_constant(
        embedding_dim, patch_size, head_count, layer_count, forward_expansion, dropout
    )
    model = genotype_to_phenotype(genotype, transformer_type, input_dim, output_dim)
    model = model.to(device=device)

    # NOTE: Choose attention
    real_input_dim = (
        input_dim // patch_size
    )  # Important: Needs to be the divided by patch size!!!
    attention = attention_type.module1(
        embedding_dim, head_count, dropout, real_input_dim, batch_size
    )
    model.swap_attention(attention)

    evaluate_transformer(model, transformer_type, evaluation_config)

    # Calculate RMSE and MAE error
    print(f"$$$ Running evaluation $$$ ")
    mae_list = []
    rmse_list = []
    for idx, example in enumerate(test_set[:evaluated_examples]):
        # print(f"$ {idx + 1}/{len(test_set)}")
        source = example["x"]
        target = example["y"]
        prediction = make_prediction(
            model,
            transformer_type,
            source,
            device,
            input_dim,
            output_dim,
            patch_size,
            split_size,
        )

        if not isinstance(prediction, np.ndarray):
            prediction = np.array(prediction)

        if not isinstance(target, np.ndarray):
            target = np.array(target)

        mae_value = mae(prediction, target)
        rmse_value = rmse(prediction, target)

        mae_list.append(mae_value)
        rmse_list.append(rmse_value)

    mae_mean = mean(mae_list)
    rmse_mean = mean(rmse_list)
    print(f"$$$ MAE: {mae_mean} $$$ ")
    print(f"$$$ RMSE: {rmse_mean} $$$ ")

    # Plot example
    # NOTE: Select examples to show
    for i in range(10):
        example = test_set[i]
        source = example["x"]
        target = example["y"]
        prediction = make_prediction(
            model,
            transformer_type,
            source,
            device,
            input_dim,
            output_dim,
            patch_size,
            split_size,
        )

        if not isinstance(prediction, np.ndarray):
            prediction = np.array(prediction)

        if not isinstance(target, np.ndarray):
            target = np.array(target)

        # Save figures
        if (
            dataset_type == DatasetType.SYNTHETIC
            and i in [0, 4]
            or dataset_type == DatasetType.LIBRA
            and i in [0, 1, 2, 6]
        ):
            no_under_transformer_type = transformer_type.replace("_", "-")
            no_under_attention_type = attention_type.replace("_", "-")
            png_path = f"figures/{run_timestamp}/{no_under_transformer_type}-{no_under_attention_type}-{dataset_type}-{forecasting_horizon}-{i}.png"
        else:
            png_path = None

        plot_prediction(source, target, prediction, png_path)

    return mae_mean, rmse_mean


# For visualization
# Synthetic: (0)
# Libra: (2)
if __name__ == "__main__":
    run_timestamp = datetime.datetime.now().strftime("%d-%m-%Y_%H:%M:%S")
    epochs = 10  # 5
    number_of_trainings = 10  # 5
    evaluated_examples = 1000  # 1000
    forecasting_horizons = [200]  # [100, 200, 400]
    attention_types_all = [
        Attention.Passthrough,
        Attention.VanillaAttention,
        Attention.LongformerAttention,
        Attention.BigBirdBlockSparseAttention,
        Attention.LongShortAttention,
        Attention.LocalitySensitiveHashingAttentionReformer,
        Attention.KmeansRoutingAttention,
        Attention.AdaptiveSpanAttention,
        Attention.LinformerAttention,
        Attention.FavorAttentionPerformer,
        Attention.RelativeMultiHeadAttentionXL,
        Attention.AttentionFreeFullAttention,
        Attention.FNetFourierTransformAttention,
        Attention.FactorizedDenseAttention,
        Attention.ConformerConvolutionAttention,
    ]
    attention_types_selected = [
        Attention.BigBirdBlockSparseAttention,
        Attention.LongformerAttention,
        Attention.ConformerConvolutionAttention,
        Attention.LocalitySensitiveHashingAttentionReformer,
        Attention.AdaptiveSpanAttention,
        Attention.RelativeMultiHeadAttentionXL,
        Attention.VanillaAttention,
    ]
    # NOTE: Select attention types
    attention_types = attention_types_selected  # attention_types_all
    transformer_types = [
        TransformerType.ENCODER_ONLY,
        TransformerType.ENCODER_ONLY_SIMPLE,
    ]
    dataset_types = [DatasetType.SYNTHETIC, DatasetType.LIBRA]
    parameters_synthetic = [[8, 50, 2, 2, 2, 0.01], [32, 5, 2, 2, 3, 0.01]]
    parameters_libra = [[32, 5, 4, 7, 6, 0.01], [64, 5, 4, 9, 7, 0.02]]

    # Also save the data in pandas df
    df = pd.DataFrame()
    for c in [
        "transformer_type",
        "parameters",
        "attention_type",
        "dataset_horizon",
        "mae",
        "rmse",
    ]:
        df[c] = ""
    df_csv_name = f"attention_{run_timestamp}.csv"

    for idx, transformer_type in enumerate(transformer_types):
        for attention_type in attention_types:
            for dataset_type in dataset_types:
                if dataset_type == DatasetType.SYNTHETIC:
                    parameters = parameters_synthetic[idx]
                else:
                    parameters = parameters_libra[idx]
                for forecasting_horizon in forecasting_horizons:
                    evaluation_params = f"{transformer_type}, {attention_type}, {dataset_type}, {forecasting_horizon}, {parameters}"
                    print(evaluation_params)

                    maes = []
                    rmses = []
                    for _ in range(number_of_trainings):
                        mae_value, rmse_value = evaluate(
                            dataset_type,
                            forecasting_horizon,
                            transformer_type,
                            attention_type,
                            parameters,
                            epochs,
                            evaluated_examples,
                        )
                        maes.append(mae_value)
                        rmses.append(rmse_value)

                        # Append data to pandas dataframe
                        df.loc[len(df.index)] = [
                            transformer_type.value,
                            parameters,
                            attention_type.value,
                            f"{dataset_type} {forecasting_horizon}",
                            mae_value,
                            rmse_value,
                        ]

                    df.to_csv(df_csv_name)

                    mae_mean = np.mean(maes)
                    rmse_mean = np.mean(rmses)
                    mae_std = np.std(maes)
                    rmse_std = np.std(rmses)
                    evaluation_result = f"mae: ({mae_mean:5f}, {mae_std:5f}), rmse: ({rmse_mean:5f}, {rmse_std:5f})"
                    print(evaluation_result)

                    with open("attention_results.txt", "a") as fa:
                        fa.write(f"{evaluation_params}\n")
                        fa.write(f"{evaluation_result}\n\n")
