from typing import List

import pmdarima as pm
from pmdarima import ARIMA
from pmdarima.arima import ADFTest

from experiments.e1.src.visualization import plot_prediction
from src.dataset.LibraDataset import LibraDataset, LibraCategory
from src.dataset.SyntheticDataset import SyntheticDataset, SeriesType

# http://alkaline-ml.com/pmdarima/modules/generated/pmdarima.arima.AutoARIMA.html


def prepare_synthetic_time_series(dataset_length: int) -> List[List[float]]:
    dataset1 = SyntheticDataset(
        SeriesType.SIMPLE,
        length=dataset_length,
    )
    dataset2 = SyntheticDataset(
        SeriesType.COMPOSED,
        length=dataset_length,
    )
    dataset3 = SyntheticDataset(
        SeriesType.COMPOSED_WITH_NOISE,
        length=dataset_length,
    )
    series = [dataset1.raw_series(), dataset2.raw_series(), dataset3.raw_series()]
    return series


def prepare_libra_category_time_series(
    libra_category: LibraCategory, min_training_examples: int, dataset_filter=None
) -> List[List[float]]:
    dataset = LibraDataset(
        [libra_category],
        dataset_filter=dataset_filter,
        min_training_examples_available=min_training_examples,
    )
    series = dataset.raw_series()
    return series


def prepare_libra_categories_time_series(
    libra_categories: List[LibraCategory],
    min_training_examples: int,
    dataset_filter=None,
) -> List[List[float]]:
    dataset = LibraDataset(
        libra_categories,
        dataset_filter=dataset_filter,
        min_training_examples_available=min_training_examples,
    )
    series = dataset.raw_series()
    return series


def is_series_stationary(series, alpha=0.05):
    adf_test = ADFTest(alpha=alpha)
    return adf_test.should_diff(series)


def arima_auto(series, max_pqd=(5, 5, 5), max_PQD=(5, 5, 5), m=12):
    seasonal = max_PQD != (0, 0, 0)
    # Fit stepwise auto-ARIMA
    arima = pm.auto_arima(
        series,
        start_p=1,
        start_q=1,
        start_d=1,
        d=None,
        max_p=max_pqd[0],
        max_q=max_pqd[1],
        max_d=max_pqd[2],
        seasonal=seasonal,
        start_P=0,
        start_Q=0,
        start_D=0,
        D=None,
        max_P=max_PQD[0],
        max_Q=max_PQD[1],
        max_D=max_PQD[2],
        m=m,
        trace=True,
        error_action="warn",
        suppress_warnings=True,
        stepwise=True,
    )
    return arima


def arima_auto_fit_and_predict(
    x, predict_tokens, max_order=(5, 5, 5), max_seasonal_order=(3, 3, 3, 12)
):
    model = arima_auto(
        x,
        max_pqd=max_order,
        max_PQD=(max_seasonal_order[0], max_seasonal_order[1], max_seasonal_order[3]),
        m=max_seasonal_order[-1],
    )
    prediction = model.predict(n_periods=predict_tokens)
    return prediction


def arima_manual(order=(3, 3, 3), seasonal_order=(3, 3, 3, 12)):
    arima = ARIMA(order=order, seasonal_order=seasonal_order)  # (p,d,q)  # (P,D,Q,s)
    return arima


def arima_manual_fit(series, order=(3, 3, 3), seasonal_order=(3, 3, 3, 12)):
    arima = ARIMA(order=order, seasonal_order=seasonal_order)  # (p,d,q)  # (P,D,Q,s)
    arima.fit(series)
    return arima


def arima_manual_fit_and_predict(
    source, predict_tokens, order=(3, 3, 3), seasonal_order=(3, 3, 3, 12)
):
    model = arima_manual_fit(source, order=order, seasonal_order=seasonal_order)
    prediction = model.predict(n_periods=predict_tokens)
    return prediction


def arima_predict(model, output_dim):
    prediction = model.predict(n_periods=output_dim)
    return prediction


def main():
    max_samples = 2000
    split_ratio = 0.8

    dataset = SyntheticDataset(SeriesType.COMPOSED_WITH_NOISE, x_length=1, y_length=1)
    raw_data = dataset.raw_series()[:max_samples]

    dataset_length = len(raw_data)
    split_index = int(dataset_length * split_ratio)
    train_set, test_set = raw_data[:split_index], raw_data[split_index:]

    arima = arima_auto(train_set, max_pqd=(3, 3, 3), max_PQD=(3, 3, 3))
    prediction = arima_predict(arima, dataset_length - split_index)

    plot_prediction(train_set, test_set, prediction)


if __name__ == "__main__":
    main()
