import datetime
from enum import Enum

import numpy as np
import pandas as pd
import torch

from experiments.e1.src.evaluation import TransformerType
from experiments.e1.src.visualization import make_prediction as make_prediction_genotype
from experiments.e4.arima.train_arima import (
    arima_manual_fit_and_predict,
    arima_auto_fit_and_predict,
)
from experiments.e4.genotype.train_genotype import DatasetType
from src.dataset.LibraDataset import LibraCategory
from src.evolutionary.dataset.bootstrap import (
    boostrap_synthetic_dataset,
    boostrap_libra_dataset,
)
from src.training.evaluation import mae, rmse
from src.utils.helpers import mean

from experiments.e4.genotype.train_genotype import train as train_genotype
from experiments.e4.rnn.train_rnn import train as train_rnn_lstm, RecurrentType
from experiments.e4.rnn.train_rnn import make_prediction as make_prediction_rnn_lstm
from experiments.e4.nbeats.train_nbeats import train as train_nbeats
from experiments.e4.nbeats.train_nbeats import make_prediction as make_prediction_nbeats
from experiments.e4.tcn.train_tcn import train as train_tcn
from experiments.e4.tcn.train_tcn import make_prediction as make_prediction_tcn
from experiments.e4.vanilla_encoder.train_vanilla_encoder import (
    train as train_vanilla_encoder,
)


class MethodType(str, Enum):
    ARIMA = "arima"
    RNN = "rnn"
    LSTM = "lstm"
    TCN = "tcn"
    NBEATS = "nbeats"
    TransformerEncoder = "vanilla_encoder"
    BestGenotype = "best_genotype"


def evaluate(
    dataset_type: DatasetType,
    libra_category: LibraCategory,
    dataset_filter: list = None,
    libra_min_training_examples: int = 100,
    libra_max_values_per_series: int = 1000,
    method_type: MethodType = MethodType.BestGenotype,
    input_dim: int = 400,
    output_dim: int = 50,
    evaluated_examples: int = 1000,
    epochs: int = 5,
    encoder_parameters: list = None,
    genotype_path: str = None,
):
    # Datasets
    batch_size = 16
    train_test_ratio = 0.8

    synthetic_length = 5000

    # Other configuration
    use_auto_arima = False
    device = "cuda" if torch.cuda.is_available() else "cpu"
    dropout = 0.1

    # Train and test sets
    if dataset_type == DatasetType.SYNTHETIC:
        train_set, test_set = boostrap_synthetic_dataset(
            input_dim,
            output_dim,
            synthetic_length,
            train_test_ratio,
        )
    elif dataset_type == DatasetType.LIBRA:
        train_set, test_set = boostrap_libra_dataset(
            input_dim,
            output_dim,
            [libra_category],
            [dataset_filter],
            libra_min_training_examples,
            libra_max_values_per_series,
            train_test_ratio,
        )
    else:
        raise ValueError("Unknown dataset type")

    print(f"# Dataset length: {len(train_set) + len(test_set)}")

    # Training
    if method_type == MethodType.ARIMA:
        pass  # ARIMA is fitted when evaluated
    elif method_type == MethodType.RNN:
        model = train_rnn_lstm(
            RecurrentType.RNN,
            batch_size,
            input_dim,
            output_dim,
            train_set,
            test_set,
            epochs,
            device,
            dropout,
        )
    elif method_type == MethodType.LSTM:
        model = train_rnn_lstm(
            RecurrentType.LSTM,
            batch_size,
            input_dim,
            output_dim,
            train_set,
            test_set,
            epochs,
            device,
            dropout,
        )
    elif method_type == MethodType.TCN:
        model = train_tcn(
            batch_size,
            input_dim,
            output_dim,
            train_set,
            test_set,
            epochs,
            device,
            dropout,
        )
    elif method_type == MethodType.NBEATS:
        model = train_nbeats(
            batch_size, input_dim, output_dim, train_set, test_set, epochs, device
        )
    elif method_type == MethodType.TransformerEncoder:
        model = train_vanilla_encoder(
            encoder_parameters,
            batch_size,
            input_dim,
            output_dim,
            train_set,
            test_set,
            epochs,
            device,
        )
    elif method_type == MethodType.BestGenotype:
        model, _, _ = train_genotype(
            batch_size,
            input_dim,
            output_dim,
            train_set,
            test_set,
            epochs,
            device,
            genotype_path,
        )

    # Calculate RMSE and MAE error
    print(f"$$$ Running evaluation $$$ ")
    mae_list = []
    rmse_list = []

    # Making predictions
    for idx, example in enumerate(test_set[:evaluated_examples]):
        # print(f"$ {idx + 1}/{len(test_set)}")
        source = example["x"]
        target = example["y"]
        source = np.nan_to_num(source)
        target = np.nan_to_num(target)

        try:
            if method_type == MethodType.ARIMA:
                print(f"$ ARIMA: {idx + 1}/{len(test_set)}")
                arima_max_order = (5, 5, 5)
                arima_order = (3, 3, 2)

                if use_auto_arima:
                    prediction = arima_auto_fit_and_predict(
                        source,
                        len(target),
                        max_order=arima_max_order,
                        max_seasonal_order=(0, 0, 0, 0),
                    )
                else:
                    prediction = arima_manual_fit_and_predict(
                        source,
                        len(target),
                        order=arima_order,
                        seasonal_order=(0, 0, 0, 0),
                    )

            elif method_type == MethodType.RNN:
                prediction = make_prediction_rnn_lstm(
                    model, source, target, input_dim, device
                )
            elif method_type == MethodType.LSTM:
                prediction = make_prediction_rnn_lstm(
                    model, source, target, input_dim, device
                )
            elif method_type == MethodType.TCN:
                predictions, _ = make_prediction_tcn(
                    model, source, target, input_dim, device
                )
                prediction = predictions[0][0]
            elif method_type == MethodType.NBEATS:
                prediction = make_prediction_nbeats(model, source)
            elif method_type == MethodType.TransformerEncoder:
                patch_size = model.patch_size()
                prediction = make_prediction_genotype(
                    model,
                    TransformerType.ENCODER_ONLY,
                    source,
                    device,
                    input_dim,
                    output_dim,
                    patch_size,
                    0,
                )
            elif method_type == MethodType.BestGenotype:
                patch_size = model.patch_size()
                prediction = make_prediction_genotype(
                    model,
                    TransformerType.ENCODER_ONLY,
                    source,
                    device,
                    input_dim,
                    output_dim,
                    patch_size,
                    0,
                )
        except Exception as e:
            print("Error: {str(e)}")
            prediction = np.zeros(len(target))

        if not isinstance(prediction, np.ndarray):
            prediction = np.array(prediction)

        if not isinstance(target, np.ndarray):
            target = np.array(target)

        mae_value = mae(prediction, target)
        rmse_value = rmse(prediction, target)

        mae_list.append(mae_value)
        rmse_list.append(rmse_value)

    mae_mean = mean(mae_list)
    rmse_mean = mean(rmse_list)
    print(f"$$$ MAE: {mae_mean} $$$ ")
    print(f"$$$ RMSE: {rmse_mean} $$$ ")

    return mae_mean, rmse_mean


def evaluate_inner_loop(
    number_of_trainings,
    method_type,
    dataset_type,
    input_dim,
    forecasting_horizon,
    evaluated_examples,
    libra_category,
    dataset_filter,
    libra_min_training_examples,
    libra_max_values_per_series,
    epochs,
    parameters,
    best_genotype_path,
):
    maes = []
    rmses = []
    for _ in range(number_of_trainings):
        mae_value, rmse_value = evaluate(
            dataset_type,
            libra_category,
            dataset_filter,
            libra_min_training_examples,
            libra_max_values_per_series,
            method_type,
            input_dim,
            forecasting_horizon,
            evaluated_examples,
            epochs,
            parameters,
            best_genotype_path,
        )
        maes.append(mae_value)
        rmses.append(rmse_value)

        # Append data to pandas dataframe
        df.loc[len(df.index)] = [
            method_type.value,
            f"{dataset_type}_{input_dim}_{forecasting_horizon}",
            libra_category,
            mae_value,
            rmse_value,
        ]

    df.to_csv(df_csv_name)

    mae_mean = np.mean(maes)
    rmse_mean = np.mean(rmses)
    mae_std = np.std(maes)
    rmse_std = np.std(rmses)
    evaluation_result = (
        f"mae: ({mae_mean:5f}, {mae_std:5f}), rmse: ({rmse_mean:5f}, {rmse_std:5f})"
    )
    print(evaluation_result)


# Method configuration
# ARIMA: p = 3, q = 3, d = 2
# RNN: hidden dimension = 64, number of layers = 2, dropout = 0.1
# LSTM: hidden dimension = 64, number of layers = 2, dropout = 0.1
# TCN: hidden dimension = 25, number of layers = 8, kernel size = 17, dropout = 0.1
# NBEATS: hidden dimension = 128, stack types = [Generic block, Generic block], Theta dimensions = [8, 8], blocks per stack = 6
# Transformer encoder: embedding dimension = 32, patch size = 5, head count = 4, layer count = 7, forward expansion = 6, dropout = 0.1
# Best evolved architecture:


if __name__ == "__main__":
    epochs = 20
    total_number_of_trainings = 10
    evaluated_examples = 1000
    method_types = [
        MethodType.ARIMA,
        MethodType.RNN,
        MethodType.LSTM,
        MethodType.TCN,
        MethodType.NBEATS,
        MethodType.TransformerEncoder,
        MethodType.BestGenotype,
    ]
    input_dims = [400]
    forecasting_horizons = [1, 20, 50]

    dataset_types = [DatasetType.LIBRA]  # [DatasetType.SYNTHETIC, DatasetType.LIBRA]
    libra_categories = [
        LibraCategory.ECONOMICS,
        LibraCategory.FINANCE,
        LibraCategory.HUMAN,
        LibraCategory.NATURE,
    ]
    libra_dataset_filters1 = [
        [71, 86, 69, 75, 70, 83, 80, 74, 81, 77],
        [63, 96, 49, 39, 71, 18, 99, 66, 82, 17],
        [86, 56, 46, 21, 67, 5, 2, 92, 68, 97],
        [0, 91, 39, 70, 23, 42, 73, 1, 53, 24],
    ]
    libra_dataset_filters2 = [  # These are the ones to use for evaluation!!!
        [68, 84, 73, 88, 87, 79, 72, 76, 26, 85],
        [61, 81, 72, 41, 57, 94, 83, 53, 19, 88],
        [99, 27, 65, 20, 17, 69, 47, 61, 13, 38],
        [33, 62, 66, 55, 46, 43, 26, 92, 18, 20],
    ]
    libra_dataset_filters = libra_dataset_filters2
    libra_min_training_examples = 100
    libra_max_values_per_series = 5000

    best_genotype_path = "genotype/randm-all-best-genotype33.json"
    encoder_parameters_synthetic = [8, 50, 2, 2, 2, 0.01]
    encoder_parameters_libra = [32, 5, 4, 7, 6, 0.01]

    # Also save the data in pandas df
    df = pd.DataFrame()
    for c in ["method_type", "dataset_input_horizon", "category", "mae", "rmse"]:
        df[c] = ""
    df_csv_name = f"methods_{datetime.datetime.now().strftime('%d-%m-%Y_%H:%M:%S')}.csv"

    for method_type in method_types:
        if method_type == MethodType.ARIMA:
            number_of_trainings = 1
        else:
            number_of_trainings = total_number_of_trainings
        for dataset_type in dataset_types:
            # Synthetic
            if dataset_type == DatasetType.SYNTHETIC:
                parameters = encoder_parameters_synthetic
                for input_dim in input_dims:
                    for forecasting_horizon in forecasting_horizons:
                        evaluation_params = f"{method_type}, {dataset_type}, {input_dim}, {forecasting_horizon}"
                        print(evaluation_params)

                        evaluate_inner_loop(
                            number_of_trainings,
                            method_type,
                            dataset_type,
                            input_dim,
                            forecasting_horizon,
                            evaluated_examples,
                            None,
                            None,
                            None,
                            None,
                            epochs,
                            parameters,
                            best_genotype_path,
                        )
            # Libra
            else:
                parameters = encoder_parameters_libra
                for idx, libra_category in enumerate(libra_categories):
                    libra_dataset_filter = libra_dataset_filters[idx]

                    for input_dim in input_dims:
                        for forecasting_horizon in forecasting_horizons:
                            evaluation_params = f"{method_type}, {dataset_type}, {libra_category}, {input_dim}, {forecasting_horizon}"
                            print(evaluation_params)

                            evaluate_inner_loop(
                                number_of_trainings,
                                method_type,
                                dataset_type,
                                input_dim,
                                forecasting_horizon,
                                evaluated_examples,
                                libra_category,
                                libra_dataset_filter,
                                libra_min_training_examples,
                                libra_max_values_per_series,
                                epochs,
                                parameters,
                                best_genotype_path,
                            )
