import random
import torch
from torch import optim, nn
from torch.utils.data import DataLoader

from experiments.e4.utils.batch import batch_iterator, create_prediction_batch
from experiments.e4.rnn.Optimization import Optimization
from experiments.e4.tcn.tcn import TCN
from experiments.e4.utils.visualisation import plot_prediction
from src.dataset.SyntheticDataset import SyntheticDataset, SeriesType
from src.utils.seed import temp_seed

# https://github.com/timeseriesAI/tsai/blob/main/nbs/113_models.TCN.ipynb
# https://timeseriesai.github.io/tsai/models.TCN.html


def train(
    batch_size, input_dim, output_dim, train_set, test_set, epochs, device, dropout
):
    model = TCN(c_in=1, c_out=output_dim, conv_dropout=dropout, fc_dropout=dropout).to(
        device=device
    )

    loss_fn = nn.SmoothL1Loss(reduction="sum")
    optimizer = optim.Adam(
        model.parameters(), betas=(0.9, 0.999), lr=1e-3, weight_decay=0
    )

    train_batches = list(
        batch_iterator(train_set, batch_size=batch_size, device=device)
    )
    test_batches = list(batch_iterator(test_set, batch_size=batch_size, device=device))

    opt = Optimization(model=model, loss_fn=loss_fn, optimizer=optimizer, device=device)
    opt.train(
        train_batches,
        test_batches,
        batch_size=batch_size,
        epochs=epochs,
        input_dim=input_dim,
    )
    # opt.plot_losses()
    return opt


def make_prediction(opt, source, target, input_dim, device):
    example_batch = create_prediction_batch(source, target, device)
    predictions, values = opt.evaluate(example_batch, batch_size=1, input_dim=input_dim)
    return predictions, values


def main():
    input_dim = 200
    output_dim = 50
    batch_size = 16
    dropout = 0.05
    epochs = 20

    device = "cuda" if torch.cuda.is_available() else "cpu"

    dataset = SyntheticDataset(
        SeriesType.COMPOSED_WITH_NOISE, x_length=input_dim, y_length=output_dim
    )
    with temp_seed(1):
        train_set, test_set = torch.utils.data.random_split(dataset, [4000, 750])

    # Shuffle datasets!
    with temp_seed(1):
        train_set = [train_set[i] for i in range(len(train_set))]
        random.shuffle(train_set)
        test_set = [test_set[i] for i in range(len(test_set))]
        random.shuffle(test_set)

    opt = train(
        batch_size, input_dim, output_dim, train_set, test_set, epochs, device, dropout
    )

    example = dataset[0]
    source = example["x"]
    target = example["y"]
    predictions, values = make_prediction(opt, source, target, input_dim, device)
    plot_prediction(example["x"], example["y"], predictions[0][0])


if __name__ == "__main__":
    main()
