import numpy as np
import pandas as pd

from experiments.shared.utils import merge_mean_std_and_stringify, merge_two_dfs


def generate_columns_definition(column_count, max_total_width_cm):
    column_with_cm = max_total_width_cm / column_count
    columns = [
        "L{" + str(round(column_with_cm, 1)) + "cm}|" for _ in range(column_count)
    ]
    definition = "{|" + "".join(columns) + "}"
    return definition


def simple_cell_print(value, precision=5):
    if isinstance(value, float):
        value = round(value, precision)
    return str(value)


def snake_case_to_first_capitalized(string):
    string = string.replace("_", " ")
    string = string.capitalize()
    return string


def remove_underscores(value):
    if isinstance(value, str):
        value = value.replace("_", " ")
    return value


def table_from_df(
    df,
    df2=None,
    df3=None,
    df4=None,
    cell_print_fn=simple_cell_print,
    max_width_cm: int = 10,
    complete_table: bool = True,
    first_column_header: str = "",
) -> str:
    column_names = df.columns.values.tolist()
    column_names.insert(0, first_column_header)

    # Complete table environment
    if complete_table:
        prefix_str = "\\begin{table}[ht]\n\\begin{center}\n\\bgroup\n\def\\arraystretch{1.2}\n\scalebox{1.0}{\n"
        postfix_str = (
            "}\n\egroup\n\caption{.}\n\label{tab:Label}\n\end{center}\n\end{table}"
        )
    else:
        prefix_str = ""
        postfix_str = ""

    # Add column definition
    column_definition = generate_columns_definition(len(column_names), max_width_cm)
    table_start_str = "\\begin{tabular}" + column_definition + "\n\hline\hline\n"
    table_end_str = "\n\hline\n\end{tabular}"

    # Add table header
    header_str = "".join(
        [
            "\\textbf{" + snake_case_to_first_capitalized(cn) + "} & "
            for cn in column_names
        ]
    )
    header_str = header_str[:-2] + "\\\\\hline\hline\n"

    # Determine max column values to highlight
    argmin_per_column = []
    for c in df.columns.values.tolist():
        values = np.array([v for v in df[c]])
        min_index = values.argmin()
        argmin_per_column.append(min_index)

    # Merge the two dfs
    if df2 is not None:
        df = df.combine(df2, merge_mean_std_and_stringify)

    # Merge more than two dfs
    if df3 is not None and df4 is not None:
        df_second = df3.combine(df4, merge_mean_std_and_stringify)
        df = df.combine(df_second, merge_two_dfs)

    # Add table rows
    rows_str_list = []
    for idx_row, row in enumerate(df.iterrows()):
        row_name = row[0]
        row_data = row[1].tolist()

        first_column_str = "\\textit{" + str(remove_underscores(row_name)) + "} & "

        other_columns_list = []
        for idx_column, v in enumerate(row_data):
            min_index = argmin_per_column[idx_column]
            if min_index == idx_row:
                column_str = "\\textbf{" + cell_print_fn(v) + "} & "
            else:
                column_str = cell_print_fn(v) + " & "
            other_columns_list.append(column_str)

        other_columns_str = "".join(other_columns_list)
        other_columns_str = other_columns_str[:-2] + "\\\\\hline\n"

        columns_str = first_column_str + other_columns_str
        rows_str_list.append(columns_str)

    rows_str = "".join(rows_str_list)

    return (
        prefix_str
        + table_start_str
        + header_str
        + rows_str
        + table_end_str
        + postfix_str
    )


if __name__ == "__main__":
    # NOTE: Select dataset
    csv_path = "../e2/results/20220525/attention_23-05-2022_23:35:22.csv"
    df = pd.read_csv(csv_path)

    table_string = table_from_df(df)
    print(table_string)
