import pandas as pd
from scipy import stats
from bioinfokit.analys import stat
import statsmodels.api as sm
from statsmodels.formula.api import ols
from statsmodels.multivariate.manova import MANOVA

# https://pythonfordatascienceorg.wordpress.com/welch-t-test-python-pandas/
# https://www.reneshbedre.com/blog/anova.html
# https://www.upgrad.com/blog/anova-two-factor-with-replication/
# https://www.reneshbedre.com/blog/manova-python.html
# Two way ANOVA with replication = there is more than one observed sample (usually what we want)
# Standard significance: p < 0.05 => alternative hypothesis applies (that's what we want)


def shapiro_wilk_normal_distribution_check(values: list):
    w, p = stats.shapiro(values)
    return w, p  # W test statistic, p-value


def bartlett_homogeneity_of_variances_check(list_of_value_lists: list[list]):
    w, p = stats.bartlett(list_of_value_lists)
    return w, p


# Determines if there is a significant difference of mean in two value lists
def welch_t_test(values1: list, values2: list):
    t, p = stats.ttest_ind(values1, values2, equal_var=False)
    return t, p  # t test statistic, p-value


# Compares if there is a significant difference in mean of three or more categories and one continuous value
# Assumptions about data:
# * Residuals (experimental error) are approximately normally distributed
# * Homogeneity of variances
# * Observations are sampled independently of each other
def anova(data_frame, formula):
    # Example formula: value ~ C(Genotype) + C(years) + C(Genotype):C(years)'
    # For one way anova: value ~ C(treatments)
    model = ols(formula, data=data_frame).fit()
    anova_table = sm.stats.anova_lm(model, typ=2)
    return anova_table


# Pairwise comparisons to determine which differences are statistically significant
# https://www.reneshbedre.com/blog/anova.html#multiple-pairwise-comparisons-post-hoc-test
def tukeys_hsd_test(df, res_var, xfac_var, anova_model):
    # Example:
    # * res_var = "value"
    # * xfac_var = "Genotype"
    # * anova_model = "value~C(Genotype)+C(years)+C(Genotype):C(years)"
    res = stat()
    res.tukey_hsd(df=df, res_var=res_var, xfac_var=xfac_var, anova_model=anova_model)
    return res.tukey_summary


# There is one independent variable with two or more groups and at least two dependent variables (e.g. mae and rmse)
def manova(df, formula):
    # Example: formula = "height + canopy_vol ~ plant_var"
    fit = MANOVA.from_formula(formula, data=df)
    return fit.mv_test()


# https://www.reneshbedre.com/blog/anova.html
def complete_anova_run(df, formula, res_var, xfac_vars):
    # ANOVA
    res = stat()
    res.anova_stat(df=df, res_var=res_var, anova_model=formula)
    anova_df = res.anova_summary

    # Tukey HSD
    res = stat()
    res.tukey_hsd(df=df, res_var=res_var, xfac_var=xfac_vars, anova_model=formula)
    tukey_df = res.tukey_summary

    # Validate: residuals (experimental error) are approximately normally distributed
    _, shapiro_pvalue = stats.shapiro(res.anova_model_out.resid)
    normally_distributed_residuals = shapiro_pvalue < 0.05

    # Validate : homogeneity of variances
    res = stat()
    res.levene(df=df, res_var=res_var, xfac_var=xfac_vars)
    levene_df = res.levene_summary
    levene_pvalue = levene_df["Value"][2]
    homogeneous_variances = levene_pvalue > 0.05

    return (
        anova_df,
        tukey_df,
        (normally_distributed_residuals, shapiro_pvalue),
        (homogeneous_variances, levene_pvalue),
    )


if __name__ == "__main__":
    # NOTE: Select dataset
    csv_path = "../e2/attention2.csv"
    df = pd.read_csv(csv_path)
    # NOTE: Filter out columns that are interesting
    filtered_df = df[["attention_type", "dataset_horizon", "mae"]]

    # NOTE: Define the formula
    # https://www.reneshbedre.com/blog/anova.html
    formula = "mae ~ C(attention_type) + C(dataset_horizon) + C(attention_type):C(dataset_horizon)"

    anova_df, tukey_df, normality, homogeneity = complete_anova_run(
        filtered_df,
        formula,
        "mae",
        ["attention_type", "dataset_horizon"],
    )
    print(anova_df)
    print(tukey_df)
