import unittest
from itertools import combinations
from typing import List

from src.evolutionary.genotype.modules.registries.vanilla_modules import (
    modules as vanilla_modules,
)
from src.evolutionary.core.population import (
    initialize_empty_population,
    randomize_population,
)
from src.evolutionary.genotype.Genotype import Genotype
from src.evolutionary.genotype.GenotypeCell import GenotypeCell, GenotypeModuleBranch
from src.evolutionary.genotype.GenotypeLayer import GenotypeLayer
from src.evolutionary.genotype.GenotypeModule import GenotypeModule
from src.evolutionary.genotype.parameters.IParameter import IParameter
from src.utils.helpers import mean


def argmax(iterable):
    return max(enumerate(iterable), key=lambda x: x[1])[0]


def argmin(iterable):
    return min(enumerate(iterable), key=lambda x: x[1])[0]


def parameter_difference_factor(
    parameter1: IParameter, parameter2: IParameter, eps=1e-5
) -> float:
    if parameter1.get_type() != parameter2.get_type():
        return 1.0

    param1_allowed_values = parameter1.get_allowed_values()
    param2_allowed_values = parameter2.get_allowed_values()
    if param1_allowed_values != param2_allowed_values:
        return 1.0

    param1_value = parameter1.get_value()
    param2_value = parameter2.get_value()
    parameter1_index = param1_allowed_values.index(param1_value)
    parameter2_index = param2_allowed_values.index(param2_value)

    param_diff_factor = abs(parameter1_index - parameter2_index) / (
        len(param1_allowed_values) - 1 + eps
    )
    return param_diff_factor


def module_difference_factor(
    module1: GenotypeModule, module2: GenotypeModule, eps=1e-5
) -> float:
    if module1.get_category() != module2.get_category():
        return 1.0
    # If a module is from the same category but not the same module, return 0.75
    if module1.get_module_path() != module2.get_module_path():
        return 0.75

    module1_parameters = module1.get_parameters()
    module2_parameters = module2.get_parameters()
    if len(module1_parameters) == 0 or len(module2_parameters) == 0:
        return 0.0

    param_diff_factors = []
    for idx in range(len(module1_parameters)):
        module1_param = module1_parameters[idx]
        module2_param = module2_parameters[idx]
        diff_factor = parameter_difference_factor(module1_param, module2_param, eps)
        param_diff_factors.append(diff_factor)

    # Compensate for the "same category, different module" case from above
    module_diff_factor = mean(param_diff_factors) * 0.75
    return module_diff_factor


def branch_difference_factor(
    cell1: GenotypeCell, cell2: GenotypeCell, branch: GenotypeModuleBranch, eps=1e-5
) -> float:
    cell1_branch_module_count = cell1.module_count(branch)
    cell2_branch_module_count = cell2.module_count(branch)
    min_number_of_modules = min(cell1_branch_module_count, cell2_branch_module_count)

    module_diff_factors = []
    for idx in range(min_number_of_modules):
        module1 = cell1.get_module(branch, idx)
        module2 = cell2.get_module(branch, idx)
        diff_factor = module_difference_factor(module1, module2, eps)
        module_diff_factors.append(diff_factor)

    mean_diff_factor = mean(module_diff_factors) if len(module_diff_factors) > 0 else 0
    module_count_difference = abs(
        cell1_branch_module_count - cell2_branch_module_count
    ) / (max(cell1_branch_module_count, cell2_branch_module_count) + eps)
    branch_diff_factor = mean([mean_diff_factor, module_count_difference])
    return branch_diff_factor


def cell_difference_factor(cell1: GenotypeCell, cell2: GenotypeCell, eps=1e-5) -> float:
    cell1_left_branch_module_count = cell1.module_count(GenotypeModuleBranch.LEFT)
    cell2_left_branch_module_count = cell2.module_count(GenotypeModuleBranch.LEFT)
    left_branch_min_number_of_modules = min(
        cell1_left_branch_module_count, cell2_left_branch_module_count
    )

    cell1_right_branch_module_count = cell1.module_count(GenotypeModuleBranch.RIGHT)
    cell2_right_branch_module_count = cell2.module_count(GenotypeModuleBranch.RIGHT)
    right_branch_min_number_of_modules = min(
        cell1_right_branch_module_count, cell2_right_branch_module_count
    )

    if (
        left_branch_min_number_of_modules == 0
        and right_branch_min_number_of_modules == 0
    ):
        return 0.0

    left_branch_diff_factor = branch_difference_factor(
        cell1, cell2, GenotypeModuleBranch.LEFT, eps
    )
    right_branch_diff_factor = branch_difference_factor(
        cell1, cell2, GenotypeModuleBranch.RIGHT, eps
    )
    cell_diff_factor = mean([left_branch_diff_factor, right_branch_diff_factor])
    return cell_diff_factor


def layer_difference_factor(
    layer1: GenotypeLayer, layer2: GenotypeLayer, eps=1e-5
) -> float:
    layer1_cells = layer1.cells()
    layer2_cells = layer2.cells()
    min_number_of_cells = min(len(layer1_cells), len(layer2_cells))

    if min_number_of_cells == 0:
        return 0.0

    cell_diff_factors = []
    for idx in range(min_number_of_cells):
        cell1 = layer1_cells[idx]
        cell2 = layer2_cells[idx]

        # Consider the difference between inputs of the compared cells
        left_cell_input_index_diff_factor = abs(cell1[0] - cell2[0]) / (
            max(len(layer1_cells), len(layer2_cells)) + eps
        )
        right_cell_input_index_diff_factor = abs(cell1[1] - cell2[1]) / (
            max(len(layer1_cells), len(layer2_cells)) + eps
        )
        index_diff_factor = mean(
            [left_cell_input_index_diff_factor, right_cell_input_index_diff_factor]
        )

        cell_diff_factor = cell_difference_factor(cell1[2], cell2[2], eps)
        cell_diff_factors.append(mean([index_diff_factor, cell_diff_factor]))

    layer_diff_factor = mean(cell_diff_factors)
    return layer_diff_factor


def genotype_difference_factor(
    genotype1: Genotype, genotype2: Genotype, eps=1e-5
) -> float:
    # Genotype hyperparameters
    genotype1_params = genotype1.parameters()
    genotype2_params = genotype2.parameters()

    param_diff_factors = []
    for idx in range(len(genotype1_params)):
        param1 = genotype1_params[idx]
        param2 = genotype2_params[idx]
        diff_factor = parameter_difference_factor(param1, param2, eps)
        param_diff_factors.append(diff_factor)

    # Genotype layer
    # There is only one Genotype layer per Genotype stacked k times as defined by hyperparameters
    genotype1_layer = genotype1.get_layer(0)
    genotype2_layer = genotype2.get_layer(0)
    layer_diff_factor = layer_difference_factor(genotype1_layer, genotype2_layer, eps)

    genotype_diff_factor = mean([layer_diff_factor] + param_diff_factors)
    return genotype_diff_factor


def population_difference_factor(genotypes: List[Genotype]):
    pairs = list(combinations(genotypes, 2))
    factors = [genotype_difference_factor(g1, g2) for g1, g2 in pairs]
    factor_mean = mean(factors)
    return factor_mean


class TestGenotypeDifferenceFactor(unittest.TestCase):
    def test_empty_population(self):
        genotype1 = Genotype()
        genotype2 = Genotype()
        diff_factor = genotype_difference_factor(genotype1, genotype2)
        self.assertEqual(diff_factor, 0.0)

    def test_no_difference(self):
        genotype = Genotype()
        genotypes = initialize_empty_population(genotype, 2)
        diff_factor = genotype_difference_factor(genotypes[0], genotypes[1])
        self.assertEqual(diff_factor, 0.0)

    def test_some_difference(self):
        genotype = Genotype()
        genotypes = initialize_empty_population(genotype, 2)
        genotype1 = genotypes[0]
        genotype2 = genotypes[1]
        randomize_population([genotype2], vanilla_modules, False, 0, 10, 20, 50)
        diff_factor1 = genotype_difference_factor(genotype1, genotype2)
        self.assertEqual(diff_factor1 > 0.0, True)
