from typing import Optional

import torch
from torch import nn

# Attention free attention based on AFT Full
# Adapted from: https://github.com/idiap/fast-transformers/blob/master/fast_transformers/attention/aft_attention.py
# Paper: An Attention Free Transformer


class AttentionFreeFullAttention(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        max_sequence_length: int = 1024,
        aft_parameterization=64,
    ):
        super(AttentionFreeFullAttention, self).__init__()
        self.attention = AttentionFreeFullAttention1(
            embedding_dim, head_count, max_sequence_length, aft_parameterization
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class AttentionFreeFullAttention1(nn.Module):
    """Implement the "full" attention proposed An Attention Free Transformer

    AFT attention uses only element wise operations to generate the new values
    given the queries, keys and values. The full AFT is computed as follows:

        v' = sigmoid(q) * (softmax(K + w_q, dim=1) * V).sum(dim=1)

    where q is a single query, K and V are the key and value matrices and w_q
    is a learnable vector for the given query position

    Args:
        max_sequence_length: defines the maximum acceptable sequence
                             length in order to allocate the learnable
                             parameters
        aft_parameterization: defines the dimensionality of the low rank
                              parameterization for the position bias
    """

    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        max_sequence_length: int = 1024,
        aft_parameterization=64,
    ):
        super(AttentionFreeFullAttention1, self).__init__()
        self.embedding_dim = embedding_dim
        self.head_count = head_count

        # Dimension of queries, keys and values
        # We assume d_v always equals d_k
        assert (
            embedding_dim % head_count == 0
        ), "'embedding_dim' needs to be divisible by 'head_count'"
        self.d_k = embedding_dim // head_count

        self.queries = nn.Linear(embedding_dim, embedding_dim)
        self.keys = nn.Linear(embedding_dim, embedding_dim)
        self.values = nn.Linear(embedding_dim, embedding_dim)
        self.output = nn.Linear(embedding_dim, embedding_dim)

        self.u = torch.nn.Parameter(
            torch.randn(max_sequence_length, aft_parameterization) * 0.01
        )
        self.v = torch.nn.Parameter(
            torch.randn(max_sequence_length, aft_parameterization) * 0.01
        )

    @staticmethod
    def create_local_mask(sequence_length, local_window_size):
        local_mask = torch.ones(sequence_length, sequence_length)
        local_mask = torch.tril(local_mask, local_window_size - 1)
        local_mask = torch.triu(local_mask, -(local_window_size - 1))
        return local_mask

    # NOTE: Masking is not supported
    def forward(
        self,
        queries: torch.Tensor,
        keys: torch.Tensor,
        values: torch.Tensor,
        _mask: Optional[torch.Tensor] = None,
    ):
        batch_size = queries.size(0)
        queries = (
            self.queries(queries)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        keys = (
            self.keys(keys)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        values = (
            self.values(values)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )

        N, L, H, E = queries.shape
        _, S, _, D = values.shape
        if E != D:
            raise ValueError(
                (
                    "AFT requires that queries, keys and values "
                    "have the same dimensionality"
                )
            )

        # Make a gate out of Q and permute the dimensions of K and V
        Q = torch.sigmoid(queries)
        K = keys.permute(0, 2, 3, 1).contiguous()
        V = values.permute(0, 2, 3, 1).contiguous()

        # Compute the position bias and add it to the keys
        w = self.u[:L].mm(self.v[:S].t())
        K = K[:, None, :, :, :] * w[None, :, None, None, :]

        # Compute the new values
        K = torch.softmax(K, dim=-1)
        V = Q * torch.einsum("nlhds,nhds->nlhd", K, V)

        output = (
            V.contiguous()
            .transpose(1, 2)
            .contiguous()
            .view(batch_size, -1, self.head_count * self.d_k)
        )
        return self.output(output)
