from typing import Optional

import torch
from torch import nn

# Attention free attention based on AFT Simple
# Adapted from: https://github.com/idiap/fast-transformers/blob/master/fast_transformers/attention/aft_attention.py
# Paper: An Attention Free Transformer


class AttentionFreeSimpleAttention(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
    ):
        super(AttentionFreeSimpleAttention, self).__init__()
        self.attention = AttentionFreeSimpleAttention1(embedding_dim, head_count)

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class AttentionFreeSimpleAttention1(nn.Module):
    """Implement the "simple" attention proposed An Attention Free Transformer"""

    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
    ):
        super(AttentionFreeSimpleAttention1, self).__init__()
        self.embedding_dim = embedding_dim
        self.head_count = head_count

        # Dimension of queries, keys and values
        # We assume d_v always equals d_k
        assert (
            embedding_dim % head_count == 0
        ), "'embedding_dim' needs to be divisible by 'head_count'"
        self.d_k = embedding_dim // head_count

        self.queries = nn.Linear(embedding_dim, embedding_dim)
        self.keys = nn.Linear(embedding_dim, embedding_dim)
        self.values = nn.Linear(embedding_dim, embedding_dim)
        self.output = nn.Linear(embedding_dim, embedding_dim)

    @staticmethod
    def create_local_mask(sequence_length, local_window_size):
        local_mask = torch.ones(sequence_length, sequence_length)
        local_mask = torch.tril(local_mask, local_window_size - 1)
        local_mask = torch.triu(local_mask, -(local_window_size - 1))
        return local_mask

    # NOTE: Masking is not supported
    def forward(
        self,
        queries: torch.Tensor,
        keys: torch.Tensor,
        values: torch.Tensor,
        _mask: Optional[torch.Tensor] = None,
    ):
        batch_size = queries.size(0)
        queries = (
            self.queries(queries)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        keys = (
            self.keys(keys)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        values = (
            self.values(values)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )

        # Extract some shapes
        N, L, H, E = queries.shape
        _, S, _, D = values.shape
        if E != D:
            raise ValueError(
                (
                    "AFT requires that queries, keys and values "
                    "have the same dimensionality"
                )
            )

        # Make a gate out of queries
        queries = torch.sigmoid(queries)
        keys = torch.softmax(keys, dim=1)
        values = queries * (keys * values).sum(dim=1, keepdim=True)

        output = (
            values.contiguous()
            .transpose(1, 2)
            .contiguous()
            .view(batch_size, -1, self.head_count * self.d_k)
        )
        return self.output(output)
