import torch
from torch import nn
from torch.nn import functional as functional

# Factorized Dense Synthesizer
# Adapted from: https://github.com/10-zin/Synthesizer/blob/master/synth/synthesizer/modules.py
# Paper: Synthesizer: Rethinking Self-Attention for Transformer Models


class FactorizedDenseAttentionSynthesizer(nn.Module):
    def __init__(
        self, embedding_dim, head_count, dropout=0.0, max_sequence_length=1024, factor=1
    ):
        super(FactorizedDenseAttentionSynthesizer, self).__init__()
        self.attention = FactorizedDenseAttentionSynthesizer1(
            embedding_dim, head_count, dropout, max_sequence_length, factor
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class FactorizedDenseAttentionSynthesizer1(nn.Module):
    """Factorized dense attention
    Args:
        factor: factorization factor
        max_sequence_length: defines the maximum acceptable sequence length
        dropout: dropout probability at the module output
    """

    def __init__(
        self, embedding_dim, head_count, dropout=0.0, max_sequence_length=1024, factor=1
    ):
        super(FactorizedDenseAttentionSynthesizer1, self).__init__()
        self.embedding_dim = embedding_dim
        self.max_sequence_length = max_sequence_length
        self.factor = factor

        self.f_a = nn.Linear(self.embedding_dim, factor)
        self.f_b = nn.Linear(self.embedding_dim, int(max_sequence_length / factor))
        self.dropout = nn.Dropout(dropout)

    def forward(self, q, k, v, mask=None):
        h_a = torch.repeat_interleave(
            self.f_a(q), int(self.max_sequence_length / self.factor), -1
        )
        h_b = torch.repeat_interleave(self.f_b(q), self.factor, -1)
        dense_attention = torch.matmul(h_a, h_b.transpose(1, 2))

        if mask is not None:
            dense_attention = dense_attention.masked_fill(mask == 0, -1e9)

        dense_attention = self.dropout(functional.softmax(dense_attention, dim=-1))
        output = torch.matmul(dense_attention, v)

        return output
