import math
from typing import Tuple, Optional

import torch
from torch import autocast, nn

from src.components.attention.utils.softmax import (
    SMHyperbolic,
    SMReg,
    SMOrf,
    FeatureMapType,
)

# Fast Attention Via positive Orthogonal Random features
# Adapted from: https://github.com/facebookresearch/adaptive-span/blob/main/adaptive_span.py
# Paper: Rethinking Attention with Performers


class FavorAttentionPerformer(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        dropout: float = 0.0,
        causal: bool = False,
        dim_features: Optional[int] = None,
        iter_before_redraw: Optional[int] = None,
        feature_map_type: FeatureMapType = FeatureMapType.SMReg,
        normalize_inputs: bool = True,
    ):
        super(FavorAttentionPerformer, self).__init__()
        self.attention = FavorAttentionPerformer1(
            embedding_dim,
            head_count,
            dropout,
            causal,
            dim_features,
            iter_before_redraw,
            feature_map_type,
            normalize_inputs,
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class FavorAttentionPerformer1(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        dropout: float = 0.0,
        causal: bool = False,
        dim_features: Optional[int] = None,
        iter_before_redraw: Optional[int] = None,
        feature_map_type: FeatureMapType = FeatureMapType.SMReg,
        normalize_inputs: bool = True,
    ):
        """
        Kernelized attention, as proposed in Performers
        FAVOR => "Fast Attention Via positive Orthogonal Random features"
        Args:
            dropout (float): the probability of an output to be randomly dropped at training time
            dim_features (int): the dimension of the random features space
            iter_before_redraw (int): the number of steps (forward calls) before a redraw of the features
            feature_map_type (FeatureMapType): the type of feature map being used,
            for instance orthogonal random features.
            normalize_inputs (bool): you guessed it
        """
        super(FavorAttentionPerformer1, self).__init__()

        self.causal = causal
        self.iter_before_redraw = (
            (2 * iter_before_redraw)
            if iter_before_redraw is not None
            else iter_before_redraw
        )  # This will be used for both key and query
        self.normalize_inputs = normalize_inputs
        self.feature_map_type = feature_map_type
        self.attn_drop = nn.Dropout(dropout, inplace=True)

        # Setup dimension-dependent variables
        # Reasonable dimension default
        if dim_features is None:
            assert head_count is not None, "dim_features or dim_head needs to be passed"

            # FAVOR: Automatically setting the random mapping dimension to {self.dim_features} from {head_count}
            self.dim_features = math.ceil(head_count * (1 + math.log2(head_count)))
            self.dim_features = 2 * (
                self.dim_features // 2
            )  # needs to be even for some variants
        else:
            self.dim_features = dim_features

        feature_map_constructor = {
            FeatureMapType.SMHyp: SMHyperbolic,
            FeatureMapType.SMReg: SMReg,
            FeatureMapType.SMOrf: SMOrf,
        }[self.feature_map_type]

        feature_settings = {
            "dim_features": self.dim_features,
            "iter_before_redraw": self.iter_before_redraw,
            "normalize_inputs": self.normalize_inputs,
        }

        self.feature_map: FeatureMap = feature_map_constructor(**feature_settings)  # type: ignore

    @staticmethod
    def _maybe_promote(x: torch.Tensor) -> torch.Tensor:
        # Only promote fp16 buffers, bfloat16 would be fine for instance
        return x.float() if x.dtype == torch.float16 else x

    @staticmethod
    def _causal_attention(
        k_prime: torch.Tensor, q_prime: torch.Tensor, v: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # Algorithm 1 in the paper
        ref_v = torch.ones_like(v.unsqueeze(2))  # BATCH x SEQ x 1 x EMB
        Gps = k_prime.unsqueeze(3) * v.unsqueeze(2)
        Grenorm = k_prime.unsqueeze(3) * ref_v

        # Consolidate against the feature dimension
        att_raw = torch.einsum("bcfe,bcf->bce", Gps, q_prime)
        att_norm = torch.einsum("bcfe,bcf->bce", Grenorm, q_prime)

        # Cumulative sum over the sequence
        att_raw = att_raw.cumsum(2)
        att_norm = att_norm.cumsum(2)

        return att_raw, att_norm

    # Mask is not supported
    def forward(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        *_,
        **__,
    ):

        # Project key and queries onto the feature map space
        k_prime = self.feature_map(k)
        q_prime = self.feature_map(q)

        device_type = str(q.device).split(":")[0]
        with autocast(enabled=False, device_type=device_type):
            # The softmax kernel approximation for Favor will easily overflow
            # Force the computations here to stay in fp32 for numerical stability
            # Note that the dimensions are vastly reduced when compared to scaled_dot_product
            k_prime = self._maybe_promote(k_prime)
            q_prime = self._maybe_promote(q_prime)
            v = self._maybe_promote(v)

            if not self.causal:
                att_normalization = q_prime @ (
                    k_prime.transpose(-2, -1) @ torch.ones_like(v)
                )
                att_raw = q_prime @ (k_prime.transpose(-2, -1) @ v)
            else:
                # Actually compute attention
                att_raw, att_normalization = self._causal_attention(k_prime, q_prime, v)

            # Normalize
            attention = att_raw / att_normalization

        if self.attn_drop is not None:
            attention = self.attn_drop(attention)

        return attention
