import torch
from torch import nn

# Attention approximation using Lambda layers
# Adapted from: https://github.com/facebookresearch/xformers/blob/main/xformers/components/attention/lambda_layer.py
# Paper: LambdaNetworks: Modeling long-range Interactions without Attention


class LambdaLayerAttention(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        input_dim: int,
        dropout: float = 0.0,
    ):
        super(LambdaLayerAttention, self).__init__()
        self.attention = LambdaLayerAttention1(
            embedding_dim, head_count, input_dim, dropout
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


# Note: Currently this class can't be used in decoder!
class LambdaLayerAttention1(nn.Module):
    """Lambda layer attention
    Args:
        dropout: dropout probability at the module output
    """

    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        input_dim: int,
        dropout: float = 0.0,
    ):
        super().__init__()
        self.embedding_dim = embedding_dim

        self.relative_positional_embedding = torch.nn.Parameter(
            torch.randn(2 * input_dim - 1, int(embedding_dim))
        )
        self.relative_positions = calculate_relative_positions(input_dim)

        self.dropout = torch.nn.Dropout(dropout, inplace=True)

    # Masking is not supported
    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        _mask: torch.Tensor,
    ):
        content_lambda = torch.einsum("bnk,bnv->bkv", torch.softmax(key, dim=-1), value)
        content_output = torch.einsum("bnk,bkv->bnv", query, content_lambda)

        rel_pos_emb = self.relative_positional_embedding[self.relative_positions]

        # Handle real sequence length being possibly smaller
        seq_len = query.shape[1]
        rel_pos_emb = rel_pos_emb[:seq_len, :seq_len, :]

        # Compute the position lambda for every possible combination in one go
        # then compute the position related contribution
        position_lambdas = torch.einsum(
            "mnk,bnv->bnkv", rel_pos_emb, value
        )  # One lambda per position

        unsqueezed_query = query.unsqueeze(2)
        position_output = (unsqueezed_query @ position_lambdas).squeeze()
        attention = content_output + position_output

        attention = self.dropout(attention)
        return attention


def calculate_relative_positions(n: int):
    # Adapted from LucidRains
    # https://github.com/lucidrains/lambda-networks/blob/main/lambda_networks/lambda_networks.py
    relative_positions = torch.arange(n)[None, :] - torch.arange(n)[:, None]  # (n, n)
    relative_positions += n - 1  # Shift value range from (-n+1, n-1) to (0, 2n-2)
    return relative_positions
