import numpy as np
import torch
from einops import rearrange
from torch import nn

# Linformer
# Adapted from: https://github.com/The-AI-Summer/self-attention-cv/blob/main/self_attention_cv/linformer/linformer.py
# Paper: Linformer: Self-Attention with Linear Complexity


class LinformerAttention(nn.Module):
    def __init__(
        self,
        embedding_dim,
        head_count,
        input_dim,
        projection_dim=12,
        trainable_proj=True,
    ):
        super(LinformerAttention, self).__init__()
        self.attention = LinformerAttention1(
            embedding_dim, head_count, input_dim, projection_dim, trainable_proj
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class LinformerAttention1(nn.Module):
    """Linformer attention
    Args:
        projection_dim: projection dimension of the Linformer
        trainable_proj: if the projection matrix E matrix is trainable
    """

    def __init__(
        self,
        embedding_dim,
        head_count,
        input_dim,
        projection_dim=12,
        trainable_proj=True,
    ):
        super(LinformerAttention1, self).__init__()
        self.dim_head = int(embedding_dim / head_count)
        self.heads = head_count

        self.to_qvk = nn.Linear(embedding_dim, embedding_dim * 3, bias=False)
        self.W_0 = nn.Linear(embedding_dim, embedding_dim, bias=False)
        self.scale_factor = self.dim_head**-0.5

        proj_shape = (input_dim, projection_dim)
        self.E = torch.nn.Parameter(
            torch.randn(proj_shape), requires_grad=trainable_proj
        )
        self.k = proj_shape[1]

    # Mask is not supported
    def forward(self, x, _key, _value, _mask):
        assert x.dim() == 3
        assert x.shape[1] == self.E.shape[0], (
            f"{x.shape[1]} Token in the input sequence while"
            f" {self.E.shape[0]} were provided in the E proj matrix"
        )

        qkv = self.to_qvk(x)  # (batch, tokens, dim*3*heads)
        q, k, v = tuple(rearrange(qkv, "b t (d k h ) -> k b h t d ", k=3, h=self.heads))

        v, k = project_vk_linformer(v, k, self.E)

        out = compute_mhsa(q, k, v, scale_factor=self.scale_factor)
        # Re-compose: merge heads with dim_head

        out = rearrange(out, "b h i d -> b i (h d)")
        # Apply final linear transformation layer
        return self.W_0(out)


def project_vk_linformer(v, k, E):
    # Project k,v
    v = torch.einsum("b h j d, jk -> b h k d", v, E)
    k = torch.einsum("b h j d, jk -> b h k d", k, E)
    return v, k


def compute_mhsa(q, k, v, scale_factor=1.0, mask=None):
    # Resulted shape will be: (batch, heads, tokens, tokens)
    scaled_dot_prod = torch.einsum("... i d , ... j d -> ... i j", q, k) * scale_factor

    if mask is not None:
        assert mask.shape == scaled_dot_prod.shape[2:]
        scaled_dot_prod = scaled_dot_prod.masked_fill(mask, -np.inf)

    attention = torch.softmax(scaled_dot_prod, dim=-1)
    # Calc result per head
    return torch.einsum("... i j , ... j d -> ... i d", attention, v)
