import functools
from math import ceil, gcd

import torch
from einops import rearrange, repeat
from torch import einsum, nn
import torch.nn.functional as functional

# Long-Short Attention
# Adapted from: https://github.com/lucidrains/long-short-transformer/blob/main/long_short_transformer/long_short_transformer.py
# Paper: Long-Short Transformer: Efficient Transformers for Language and Vision


class LongShortAttention(nn.Module):
    def __init__(
        self,
        embedding_dim,
        input_dim,
        head_count,
        dropout=0.0,
        causal=False,
        window_size=None,  # 128
        window_size_divider=4,
        segment_size=16,
        r=1,
    ):
        super(LongShortAttention, self).__init__()
        self.attention = LongShortAttention1(
            embedding_dim,
            input_dim,
            head_count,
            dropout,
            causal,
            window_size,
            window_size_divider,
            segment_size,
            r,
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class LongShortAttention1(nn.Module):
    """Long-short attention
    Args:
        window_size: attention sliding window width
        segment_size: input sequence is divided into disjoint segments for efficiency reasons
        dropout: dropout probability at the module output
    """

    def __init__(
        self,
        embedding_dim,
        input_dim,
        head_count,
        dropout=0.0,
        causal=False,
        window_size=None,  # 128
        window_size_divider=4,
        segment_size=16,
        r=1,
    ):
        super().__init__()
        assert not (
            causal and r >= segment_size
        ), "r should be less than segment size, if autoregressive"

        # Dimension of queries, keys and values
        # We assume d_v always equals d_k
        assert (
            embedding_dim % head_count == 0
        ), "'embedding_dim' needs to be divisible by 'heads'"
        self.dim_head = embedding_dim // head_count

        self.scale = self.dim_head**-0.5
        self.heads = head_count
        self.causal = causal

        if window_size is None:
            window_size = input_dim // window_size_divider

        self.window_size = window_size
        self.segment_size = segment_size
        self.pad_to_multiple = (
            window_size if not causal else lcm(window_size, segment_size)
        )

        self.to_dynamic_proj = nn.Linear(self.dim_head, r, bias=False)
        self.local_norm = nn.LayerNorm(self.dim_head)
        self.global_norm = nn.LayerNorm(self.dim_head)

        self.attn_dropout = nn.Dropout(dropout)

        inner_dim = head_count * self.dim_head
        self.to_q = nn.Linear(embedding_dim, inner_dim, bias=False)
        self.to_kv = nn.Linear(embedding_dim, inner_dim, bias=False)
        self.to_out = nn.Linear(inner_dim, embedding_dim)

    def forward(self, q, k, _v, mask=None):
        b, n, *_, h, device, causal, w, s = (
            *q.shape,
            self.heads,
            q.device,
            self.causal,
            self.window_size,
            self.segment_size,
        )
        if mask is not None:
            mask = mask.squeeze(-1)

        # pad input sequence to multiples of window size (or window size and segment length if causal)

        q = pad_to_multiple(q, self.pad_to_multiple, dim=-2, value=0.0)
        k = pad_to_multiple(k, self.pad_to_multiple, dim=-2, value=0.0)

        # derive from variables

        padded_len = q.shape[-2]
        windows = padded_len // w
        is_padded = padded_len != n

        mask_value = -torch.finfo(q.dtype).max

        # handle mask if padding was needed and mask was not given

        if is_padded:
            mask = default(mask, torch.ones((b, n), device=device).bool())
            mask = pad_to_multiple(mask, w, dim=-1, value=False)

        # get queries, keys, values

        qkv = (self.to_q(q), self.to_kv(k))

        # get sequence range, for calculating mask

        seq_range = torch.arange(padded_len, device=device)

        # split heads

        q, kv = map(lambda t: rearrange(t, "b n (h d) -> (b h) n d", h=h), qkv)

        # scale queries

        q = q * self.scale

        # get local queries and keys similarity scores

        window_fn = lambda t: rearrange(t, "b (w n) d -> b w n d", n=w)
        lq, lkv = map(window_fn, (q, kv))

        lookaround_kwargs = {"backward": 1, "forward": (0 if causal else 1)}
        lkv = look_around(lkv, **lookaround_kwargs)

        lkv = self.local_norm(lkv)
        lsim = einsum("b w i d, b w j d -> b w i j", lq, lkv)

        # prepare global key / values

        if self.causal:
            # autoregressive global attention is handled in segments
            # later on, these segments are carefully masked to prevent leakage

            gkv = rearrange(kv, "b (n s) d -> b n s d", s=s)
            pkv = self.to_dynamic_proj(gkv)

            if exists(mask):
                pmask = repeat(mask, "b (n s) -> (b h) n s", s=s, h=h)
                pkv.masked_fill_(~pmask[..., None], mask_value)

            pkv = pkv.softmax(dim=-2)

            gkv = einsum("b n s d, b n s r -> b n r d", gkv, pkv)
            gkv = rearrange(gkv, "b n r d -> b (n r) d")
        else:
            # equation (3) in the paper

            pkv = self.to_dynamic_proj(kv)

            if exists(mask):
                pkv.masked_fill_(~mask[..., None], mask_value)

            pkv = pkv.softmax(dim=-2)

            gkv = einsum("b n d, b n r -> b r d", kv, pkv)

        # calculate global queries and keys similarity scores

        gkv = self.global_norm(gkv)
        gsim = einsum("b n d, b r d -> b n r", q, gkv)

        # concat values together (same as keys)

        gkv = repeat(gkv, "b r d -> b w r d", w=windows)
        v = torch.cat((gkv, lkv), dim=-2)

        # masking

        buckets, i, j = lsim.shape[-3:]

        if exists(mask):
            mask = repeat(mask, "b (w n) -> (b h) w n", n=w, h=h)
            mask = look_around(mask, pad_value=False, **lookaround_kwargs)
            mask = rearrange(mask, "b w n -> b w () n")
            lsim.masked_fill_(~mask, mask_value)

        # mask out padding

        seq_range_windowed = rearrange(seq_range, "(w n) -> () w n", w=windows)
        pad_mask = (
            look_around(seq_range_windowed, pad_value=-1, **lookaround_kwargs) == -1
        )
        lsim.masked_fill_(pad_mask[:, :, None], mask_value)

        # calculate causal masking for both global and local

        if self.causal:
            g_range = rearrange(seq_range, "(n s) -> n s", s=s)
            g_range_max = g_range.amax(dim=-1)
            g_mask = seq_range[:, None] >= g_range_max[None, :]
            g_mask = rearrange(g_mask, "i j -> () i j")
            gsim.masked_fill_(~g_mask, mask_value)

            causal_mask = torch.ones(i, j, device=device).triu_(j - i + 1).bool()
            causal_mask = repeat(causal_mask, "i j -> () u i j", u=buckets)
            lsim.masked_fill_(causal_mask, mask_value)

        # concat local and global similarities together to ready for attention

        gsim = rearrange(gsim, "b (w n) r -> b w n r", w=windows)
        sim = torch.cat((gsim, lsim), dim=-1)

        # attention

        attn = sim.softmax(dim=-1)
        attn = self.attn_dropout(attn)

        # aggregate values (same as keys, since tied) and project out

        out = einsum("b w i j, b w j d -> b w i d", attn, v)
        out = rearrange(out, "(b h) w n d -> b (w n) (h d)", h=h)
        out = out[:, :n]
        return self.to_out(out)


def exists(val):
    return val is not None


def lcm(*numbers):
    return int(functools.reduce(lambda x, y: int((x * y) / gcd(x, y)), numbers, 1))


def look_around(x, backward=1, forward=0, pad_value=-1, dim=2):
    t = x.shape[1]
    dims = (len(x.shape) - dim) * (0, 0)
    padded_x = functional.pad(x, (*dims, backward, forward), value=pad_value)
    tensors = [
        padded_x[:, ind : (ind + t), ...] for ind in range(forward + backward + 1)
    ]
    return torch.cat(tensors, dim=dim)


def default(val, d):
    return val if exists(val) else d


def pad_to_multiple(tensor, multiple, dim=-1, value=0):
    seqlen = tensor.shape[dim]
    m = seqlen / multiple

    if m.is_integer():
        return tensor

    remainder = ceil(m) * multiple - seqlen
    pad_offset = (0,) * (-1 - dim) * 2
    return functional.pad(tensor, (*pad_offset, 0, remainder), value=value)
