import torch
import torch.nn.functional as functional
from torch import nn, einsum
from einops import rearrange, repeat

# ReLA attention
# Adapted from: https://github.com/lucidrains/rela-transformer/blob/main/rela_transformer/rela_transformer.py
# Paper: Sparse Attention with Linear Units


class GatedRMSNorm(nn.Module):
    def __init__(self, dim, eps=1e-8):
        super().__init__()
        self.scale = dim**-0.5
        self.eps = eps
        self.w = nn.Parameter(torch.ones(dim))
        self.g = nn.Parameter(torch.ones(dim))

    def forward(self, x):
        norm = torch.norm(x, dim=-1, keepdim=True) * self.scale
        normed_x = x / norm.clamp(min=self.eps) * self.g
        return normed_x * (x * self.w).sigmoid()


class RectifiedLinearAttention(nn.Module):
    def __init__(
        self,
        embedding_dim,
        head_count,
        causal=False,
        num_memory_kv=0,
        relu_squared=False,
    ):
        super(RectifiedLinearAttention, self).__init__()
        self.attention = RectifiedLinearAttention1(
            embedding_dim, head_count, causal, num_memory_kv, relu_squared
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class RectifiedLinearAttention1(nn.Module):
    """Rectified linear attention
    Args:
        window_size: attention sliding window width
        segment_size:
        dropout: dropout probability at the module output
    """

    def __init__(
        self,
        embedding_dim,
        head_count,
        causal=False,
        num_memory_kv=0,
        relu_squared=False,
    ):
        super().__init__()
        self.heads = head_count

        # Dimension of queries, keys and values
        # We assume d_v always equals d_k
        assert (
            embedding_dim % head_count == 0
        ), "'embedding_dim' needs to be divisible by 'head_count'"
        self.dim_head = embedding_dim // head_count

        self.scale = self.dim_head**-0.5
        self.causal = causal
        self.relu_squared = relu_squared
        self.norm = GatedRMSNorm(embedding_dim)

        self.q = nn.Linear(embedding_dim, embedding_dim, bias=False)
        self.k = nn.Linear(embedding_dim, embedding_dim, bias=False)
        self.v = nn.Linear(embedding_dim, embedding_dim, bias=False)

        self.mem_k = nn.Parameter(torch.randn(num_memory_kv, embedding_dim))
        self.mem_v = nn.Parameter(torch.randn(num_memory_kv, embedding_dim))

        self.norm_values = GatedRMSNorm(self.dim_head)
        self.to_out = nn.Sequential(
            nn.Linear(embedding_dim, embedding_dim),
        )

    def forward(self, q, k, v, mask=None):
        b, device = q.shape[0], q.device
        h = self.heads

        if mask is not None:
            mask = mask.squeeze(-1)
        q = self.q(self.norm(q))
        k = self.q(self.norm(k))
        v = self.q(self.norm(v))

        mem_k, mem_v = map(
            lambda t: repeat(t, "n d -> b n d", b=b), (self.mem_k, self.mem_v)
        )
        k = torch.cat([mem_k, k], dim=1)
        v = torch.cat([mem_v, v], dim=1)

        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> b h n d", h=h), (q, k, v))

        q = q * self.scale
        sim = einsum("b h i d, b h j d -> b h i j", q, k)

        attn = functional.relu(sim)

        if self.relu_squared:
            attn = attn**2

        if exists(mask):
            mask = rearrange(mask, "b j -> b 1 1 j")
            attn = attn.masked_fill(~mask, 0.0)

        if self.causal:
            i, j = attn.shape[-2:]
            causal_mask = torch.ones(i, j, device=device).triu_(j - i + 1).bool()
            attn = attn.masked_fill(causal_mask, 0.0)

        out = einsum("b h i j, b h j d -> b h i d", attn, v)
        out = self.norm_values(out)

        out = rearrange(out, "b h n d -> b n (h d)")
        return self.to_out(out)


def exists(val):
    return val is not None
