import torch
from torch import nn


# Relative Multi Head Attention
# Adapted from: https://nn.labml.ai/transformers/xl/relative_mha.html
# Paper: Transformer-XL: Attentive Language Models Beyond a Fixed-Length Context


class RelativeMultiHeadAttentionXL(nn.Module):
    def __init__(self, embedding_dim: int, head_count: int):
        super(RelativeMultiHeadAttentionXL, self).__init__()
        self.attention = RelativeMultiHeadAttentionXL1(embedding_dim, head_count)

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class RelativeMultiHeadAttentionXL1(nn.Module):
    """Relative Multi-Head Attention Module"""

    def __init__(self, embedding_dim: int, head_count: int):
        super(RelativeMultiHeadAttentionXL1, self).__init__()
        self.embedding_size = embedding_dim
        self.head_count = head_count

        assert (
            embedding_dim % head_count == 0
        ), "'embedding_dim' needs to be divisible by 'head_count'"
        self.d_k = embedding_dim // head_count

        self.queries_linear = nn.Linear(embedding_dim, embedding_dim)
        self.keys_linear = nn.Linear(embedding_dim, embedding_dim)
        self.values_linear = nn.Linear(embedding_dim, embedding_dim)
        self.output_linear = nn.Linear(embedding_dim, embedding_dim)

        # Number of relative positions
        self.P = 2**12

        # Relative positional embeddings for key relative to the query
        # We need 2 * P embeddings because the keys can be before or after the query
        self.key_pos_embeddings = nn.Parameter(
            torch.zeros((self.P * 2, head_count, self.d_k)), requires_grad=True
        )
        # Relative positional embedding bias for key relative to the query
        self.key_pos_bias = nn.Parameter(
            torch.zeros((self.P * 2, head_count)), requires_grad=True
        )
        # Positional embeddings for the query is independent of the position of the query
        self.query_pos_bias = nn.Parameter(
            torch.zeros((head_count, self.d_k)), requires_grad=True
        )

    def forward(self, queries, keys, values, mask):
        # Linearly project input to obtain keys, values and queries
        batch_size = queries.size(0)
        queries = (
            self.queries_linear(queries)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        keys = (
            self.keys_linear(keys)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        values = (
            self.values_linear(values)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )

        scores = self.get_scores(queries, keys)

        if mask is not None:
            mask = mask.unsqueeze(1)
            scores = scores.masked_fill(mask == 0, -1e9)

        attention = nn.functional.softmax(scores, dim=-1)
        attention_x_values = torch.matmul(attention, values)

        # Concat the score matrices together (batch_size, -1, head_count * d_k)
        output = (
            attention_x_values.transpose(1, 2)
            .contiguous()
            .view(batch_size, -1, self.head_count * self.d_k)
        )

        return self.output_linear(output)

    def get_scores(self, query: torch.Tensor, key: torch.Tensor):
        """Get relative attention scores"""
        # Query, key and value dims: (batch_size, heads, seq_len, d_k)

        # (seq_len * 2, head_count, d_k)
        key_pos_emb = self.key_pos_embeddings[
            self.P - key.shape[2] : self.P + query.shape[2]
        ]
        # (seq_len * 2, head_count)
        key_pos_bias = self.key_pos_bias[
            self.P - key.shape[2] : self.P + query.shape[2]
        ]

        # (head_count, d_k)
        query_pos_bias = self.query_pos_bias[None, :, None, :]

        # (batch_size, heads, seq_len, d_k), (batch_size, heads, seq_len, d_k) -> (batch_size, heads, seq_len, d_k)
        ac = torch.einsum("bhid,bhjd->bhij", query + query_pos_bias, key)

        # (batch_size, heads, seq_len, d_k), (seq_len * 2, head_count, d_k) -> (batch_size, heads, seq_len, seq_len * 2)
        b = torch.einsum("bhid,jhd->bhij", query, key_pos_emb)

        # (seq_len * 2, head_count)
        d = key_pos_bias[None, :, None, :].swapaxes(1, 3)
        bd = shift_right(b + d)

        # Remove extra positions
        bd = bd[:, :, :, -key.shape[2] :]

        return ac + bd


def shift_right(x: torch.Tensor):
    """
    This method shifts i'th row of a matrix by i columns.
    If the input is `[[1, 2 ,3], [4, 5 ,6], [7, 8, 9]]`, the shifted
    result would be `[[1, 2 ,3], [0, 4, 5], [9, 0, 7]]`.
    """
    # Concatenate a column of zeros
    zero_pad = x.new_zeros(x.shape[0], 1, *x.shape[2:])
    x_padded = torch.cat([x, zero_pad], dim=1)

    # Reshape and remove excess elements from the end
    x_padded = x_padded.view(x.shape[1] + 1, x.shape[0], *x.shape[2:])
    x = x_padded[:-1].view_as(x)

    return x


if __name__ == "__main__":
    x = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    print(shift_right(x))

    x = torch.arange(1, 6)[None, :, None, None].repeat(5, 1, 1, 1)
    print(x[:, :, 0, 0])
    print(shift_right(x)[:, :, 0, 0])

    x = torch.arange(1, 6)[None, :, None, None].repeat(3, 1, 1, 1)
    print(x[:, :, 0, 0])
    print(shift_right(x)[:, :, 0, 0])
