import itertools
import math
import unittest
import random

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as functional

from src.components.attention.preprocess.QKVPreprocess import QKVPreprocess
from src.components.attention.preprocess.ScoresPreprocess import ScoresPreprocess

# Multi Head Attention
# Adapted from: http://nlp.seas.harvard.edu/2018/04/03/attention.html
# Paper: Attention Is All You Need


class ScaledDotProductMultiHeadAttention(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        q_preprocess: str = QKVPreprocess.NoPreprocess.value,
        k_preprocess: str = QKVPreprocess.NoPreprocess.value,
        v_preprocess: str = QKVPreprocess.NoPreprocess.value,
        scores_preprocess: str = ScoresPreprocess.NoPreprocess.value,
    ):
        super(ScaledDotProductMultiHeadAttention, self).__init__()
        self.attention = ScaledDotProductMultiHeadAttention1(
            embedding_dim,
            head_count,
            q_preprocess,
            k_preprocess,
            v_preprocess,
            scores_preprocess,
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class ScaledDotProductMultiHeadAttention1(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        q_preprocess: str = QKVPreprocess.NoPreprocess.value,
        k_preprocess: str = QKVPreprocess.NoPreprocess.value,
        v_preprocess: str = QKVPreprocess.NoPreprocess.value,
        scores_preprocess: str = ScoresPreprocess.NoPreprocess.value,
    ):
        super(ScaledDotProductMultiHeadAttention1, self).__init__()
        self.embedding_size = embedding_dim
        self.head_count = head_count

        # Dimension of queries, keys and values
        # We assume d_v always equals d_k
        assert (
            embedding_dim % head_count == 0
        ), "'embedding_dim' needs to be divisible by 'head_count'"
        self.d_k = embedding_dim // head_count

        self.queries_linear = nn.Linear(embedding_dim, embedding_dim)
        self.keys_linear = nn.Linear(embedding_dim, embedding_dim)
        self.values_linear = nn.Linear(embedding_dim, embedding_dim)
        self.output_linear = nn.Linear(embedding_dim, embedding_dim)

        self.q_preprocess = QKVPreprocess.from_str(q_preprocess).module(
            self.embedding_size, self.head_count, self.d_k
        )
        self.k_preprocess = QKVPreprocess.from_str(k_preprocess).module(
            self.embedding_size, self.head_count, self.d_k
        )
        self.v_preprocess = QKVPreprocess.from_str(v_preprocess).module(
            self.embedding_size, self.head_count, self.d_k
        )
        self.scores_preprocess = ScoresPreprocess.from_str(scores_preprocess).module(
            self.head_count
        )

    def fix_weights(self, value=0.1):
        nn.init.constant_(self.queries_linear.weight, value)
        nn.init.constant_(self.keys_linear.weight, value)
        nn.init.constant_(self.values_linear.weight, value)
        nn.init.constant_(self.output_linear.weight, value)

    def forward(self, queries, keys, values, mask):
        # Linearly project input to obtain keys, values and queries
        batch_size = queries.size(0)
        queries = (
            self.queries_linear(queries)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        keys = (
            self.keys_linear(keys)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        values = (
            self.values_linear(values)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )

        # QKV preprocess
        queries = self.q_preprocess(queries)
        keys = self.k_preprocess(keys)
        values = self.v_preprocess(values)

        queries_x_keys = torch.matmul(queries, keys.transpose(-2, -1))
        scores = queries_x_keys / math.sqrt(self.d_k)

        if mask is not None:
            mask = mask.unsqueeze(1)
            scores = scores.masked_fill(mask == 0, -1e9)

        # Scores preprocess
        scores = self.scores_preprocess(scores)

        attention = functional.softmax(scores, dim=-1)
        attention_x_values = torch.matmul(attention, values)

        # Concat the score matrices together (batch_size, -1, head_count * d_k)
        output = (
            attention_x_values.transpose(1, 2)
            .contiguous()
            .view(batch_size, -1, self.head_count * self.d_k)
        )
        return self.output_linear(output)


class TestMultiHeadAttention(unittest.TestCase):
    def test_query_preprocessing(self):
        batch_size = 16
        sequence_length = 200
        embedding_dims = [16, 32, 64]
        head_counts = [1, 2, 4, 8]

        qkv_preprocess = QKVPreprocess.module_list()

        for embedding_dim in embedding_dims:
            input_tensor = torch.rand((batch_size, sequence_length, embedding_dim))
            for head_count in head_counts:
                for triplets in itertools.permutations(qkv_preprocess, 3):
                    q_pre = triplets[0]
                    k_pre = triplets[1]
                    v_pre = triplets[2]
                    try:
                        attention = ScaledDotProductMultiHeadAttention1(
                            embedding_dim,
                            head_count,
                            q_pre,
                            k_pre,
                            v_pre,
                            ScoresPreprocess.NoPreprocess,
                        )
                        output_tensor = attention(input_tensor)
                    except Exception as _:
                        print(
                            f"embedding_dim: {embedding_dim}, head_count: {head_count}"
                        )
                        print(f"q_pre: {q_pre}, k_pre: {k_pre}, v_pre: {v_pre}")
                    self.assertEqual(input_tensor.shape, output_tensor.shape)

    def test_scores_preprocessing(self):
        batch_size = 16
        sequence_length = 200
        embedding_dims = [16, 32, 64]
        head_counts = [1, 2, 4, 8]

        scores_preprocess = ScoresPreprocess.module_list()

        for embedding_dim in embedding_dims:
            input_tensor = torch.rand((batch_size, sequence_length, embedding_dim))
            for head_count in head_counts:
                for preprocess in scores_preprocess:
                    try:
                        attention = ScaledDotProductMultiHeadAttention1(
                            embedding_dim,
                            head_count,
                            QKVPreprocess.NoPreprocess,
                            QKVPreprocess.NoPreprocess,
                            QKVPreprocess.NoPreprocess,
                            preprocess,
                        )
                        output_tensor = attention(input_tensor)
                    except Exception as _:
                        print(
                            f"embedding_dim: {embedding_dim}, head_count: {head_count}"
                        )
                        print(f"scores_preprocess: {scores_preprocess}")
                    self.assertEqual(input_tensor.shape, output_tensor.shape)

    def simple_test(self):
        random.seed(0)
        torch.manual_seed(0)

        embedding_dim = 4
        head_count = 2
        multi_head_attention = ScaledDotProductMultiHeadAttention(
            embedding_dim, head_count
        )
        multi_head_attention.fix_weights()

        data = np.array(
            [[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]], dtype=np.dtype("f4")
        )

        x = torch.tensor(data)
        y = multi_head_attention.forward(x, x, x, None)

        output = y.detach().numpy()
        expected = np.array(
            [
                [[0.40958, -0.12472, 0.35606, -0.00273]],
                [[1.04958, 0.51528, 0.99606, 0.63727]],
            ]
        )

        np.testing.assert_array_almost_equal(output, expected, decimal=3)
