from functools import reduce, partial
from inspect import isfunction
from operator import mul

import torch
import torch.nn.functional as functional
from torch import nn

# Sinkhorn Attention
# Adapted from: https://github.com/lucidrains/sinkhorn-transformer/blob/master/sinkhorn_transformer/sinkhorn_transformer.py
# Paper: Sparse Sinkhorn Attention


class SimpleSortNet(nn.Module):
    def __init__(
        self,
        heads,
        bucket_size,
        max_buckets,
        dim,
        non_permutative,
        temperature,
        sinkhorn_iter,
    ):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.max_buckets = max_buckets
        self.bucket_size = bucket_size
        self.non_permutative = non_permutative
        self.temperature = temperature
        self.sinkhorn_iter = sinkhorn_iter
        self.linear = nn.Parameter(torch.randn(1, heads, dim, max_buckets))
        self.act = nn.ReLU()

    def forward(self, q, k, topk=1):
        bh, t, _ = q.shape
        b = bh // self.heads
        buckets = t // self.bucket_size

        b_q, b_k = bucket(buckets, q), bucket(buckets, k)
        x = torch.cat((b_q.sum(dim=2), b_k.sum(dim=2)), dim=-1)

        W = expand_batch_and_merge_head(b, self.linear)
        R = self.act(x @ W)

        return (
            differentiable_topk(R, k=topk, temperature=self.temperature)
            if self.non_permutative
            else gumbel_sinkhorn(R, self.sinkhorn_iter, self.temperature)
        )


class AttentionSortNet(nn.Module):
    def __init__(
        self,
        heads,
        bucket_size,
        kv_bucket_size,
        dim,
        non_permutative,
        temperature,
        sinkhorn_iter,
        n_sortcut=0,
    ):
        super().__init__()
        self.heads = heads
        self.bucket_size = bucket_size
        self.kv_bucket_size = kv_bucket_size
        self.dim = dim
        self.non_permutative = non_permutative
        self.temperature = temperature
        self.sinkhorn_iter = sinkhorn_iter
        self.n_sortcut = n_sortcut

    def forward(self, q, k, topk=1):
        bh, *_, bucket_size, kv_bucket_size, device, dtype, dim = (
            *q.shape,
            self.bucket_size,
            self.kv_bucket_size,
            q.device,
            q.dtype,
            self.dim,
        )
        b = bh // self.heads

        buckets = q.shape[1] // bucket_size
        kv_buckets = k.shape[1] // kv_bucket_size

        b_q = bucket(buckets, q) if self.n_sortcut == 0 else bucket(1, q)
        b_k = bucket(kv_buckets, k)

        sq = b_q.mean(dim=2)
        sk = b_k.mean(dim=2)

        R = torch.einsum("bie,bje->bij", sq, sk).to(q) * (dim**-0.5)

        if self.non_permutative:
            k = topk if self.n_sortcut == 0 else self.n_sortcut
            return differentiable_topk(R, k=k)

        return gumbel_sinkhorn(functional.relu(R), self.sinkhorn_iter, self.temperature)


class SinkhornAttention(nn.Module):
    def __init__(
        self,
        embedding_dim,
        head_count,
        dropout=0.0,
        bucket_size=10,
        max_seq_len=1024,
        temperature=0.75,
        non_permutative=True,
        sinkhorn_iter=7,
        n_sortcut=0,
        kv_bucket_size=None,
        use_simple_sort_net=False,
        n_top_buckets=1,
    ):
        super(SinkhornAttention, self).__init__()
        self.attention = SinkhornAttention1(
            embedding_dim,
            head_count,
            dropout,
            bucket_size,
            max_seq_len,
            temperature,
            non_permutative,
            sinkhorn_iter,
            n_sortcut,
            kv_bucket_size,
            use_simple_sort_net,
            n_top_buckets,
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class SinkhornAttention1(nn.Module):
    """Sinkhorn attention
    Args:
        bucket_size: size of a bucket in elements
        temperature: temperature of the Gumbel Sinkhorn balancing mechanism
        dropout: dropout probability at the module output
    """

    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        dropout=0.0,
        bucket_size=10,
        max_seq_len=1024,
        temperature=0.75,
        non_permutative=True,
        sinkhorn_iter=7,
        n_sortcut=0,
        kv_bucket_size=None,
        use_simple_sort_net=False,
        n_top_buckets=1,
    ):
        super().__init__()

        # Dimension of queries, keys and values
        # We assume d_v always equals d_k
        assert (
            embedding_dim % head_count == 0
        ), "'embedding_dim' needs to be divisible by 'num_heads'"
        self.dim_heads = embedding_dim // head_count

        self.dim = embedding_dim
        self.heads = head_count
        self.temperature = temperature
        self.non_permutative = non_permutative
        self.sinkhorn_iter = sinkhorn_iter
        self.n_sortcut = n_sortcut

        self.bucket_size = bucket_size
        self.kv_bucket_size = default(kv_bucket_size, bucket_size)

        if use_simple_sort_net:
            self.sort_net = SimpleSortNet(
                self.heads,
                self.kv_bucket_size,
                max_seq_len // self.kv_bucket_size,
                self.dim_heads * 2,
                non_permutative=non_permutative,
                temperature=temperature,
                sinkhorn_iter=sinkhorn_iter,
            )
        else:
            self.sort_net = AttentionSortNet(
                self.heads,
                self.bucket_size,
                self.kv_bucket_size,
                self.dim_heads,
                non_permutative=non_permutative,
                temperature=temperature,
                sinkhorn_iter=sinkhorn_iter,
                n_sortcut=n_sortcut,
            )

        self.n_top_buckets = n_top_buckets
        self.dropout = nn.Dropout(dropout)

        self.queries_linear = nn.Linear(embedding_dim, embedding_dim)
        self.keys_linear = nn.Linear(embedding_dim, embedding_dim)
        self.values_linear = nn.Linear(embedding_dim, embedding_dim)
        self.output_linear = nn.Linear(embedding_dim, embedding_dim)

    def forward(self, q, k, v, mask=None):
        batch_size = q.size(0)
        q = (
            self.queries_linear(q)
            .view(batch_size, -1, self.heads, self.dim_heads)
            .transpose(1, 2)
        )
        k = (
            self.keys_linear(k)
            .view(batch_size, -1, self.heads, self.dim_heads)
            .transpose(1, 2)
        )
        v = (
            self.values_linear(v)
            .view(batch_size, -1, self.heads, self.dim_heads)
            .transpose(1, 2)
        )

        b, h, t, d_h = q.shape
        n_top = self.n_top_buckets
        bucket_size = self.bucket_size
        kv_bucket_size = self.kv_bucket_size
        device = q.device

        if mask is None:
            q_mask = kv_mask = torch.ones(q.size(0), q.size(1), 1).to(torch.bool)
        else:
            q_mask = kv_mask = mask.squeeze(-1)

        bh = b * h
        buckets = q.shape[2] // bucket_size
        kv_buckets = k.shape[2] // kv_bucket_size
        n_top = min(n_top, kv_buckets)

        merge_batch_head = partial(merge_dims, 0, 1)
        q, k, v = map(merge_batch_head, (q, k, v))

        # bucket query, key, values

        b_q = bucket(buckets, q)
        b_k, b_v = map(partial(bucket, kv_buckets), (k, v))

        bsz = b_k.shape[2]

        # calculate reordering matrix R with simple sort net

        R = self.sort_net(q, k, topk=n_top)
        R = R.type_as(q).to(q)

        # concatenate reordered buckets

        b_k_r = reorder_buckets(b_k, R)
        b_v_r = reorder_buckets(b_v, R)

        # choose the top n ranked buckets for all query buckets

        if self.n_sortcut > 0:
            b_k_r = b_k_r[:, 0 : self.n_sortcut].reshape(bh, 1, -1, d_h)
            b_v_r = b_v_r[:, 0 : self.n_sortcut].reshape(bh, 1, -1, d_h)
            b_k_r = expand_dim(b_k_r, 1, buckets)
            b_v_r = expand_dim(b_v_r, 1, buckets)
        else:
            b_k_r = b_k_r.reshape(bh, buckets, -1, d_h)
            b_v_r = b_k_r.reshape(bh, buckets, -1, d_h)

        b_k = torch.cat((b_k_r, b_k), dim=2) if buckets == kv_buckets else b_k_r
        b_v = torch.cat((b_v_r, b_v), dim=2) if buckets == kv_buckets else b_v_r

        dots = torch.einsum("buie,buje->buij", b_q, b_k) * (d_h**-0.5)

        # mask
        mask_value = max_neg_value(dots)

        if not all_none(q_mask, kv_mask):
            q_mask = default(q_mask, lambda: torch.ones((b, t), device=device).bool())
            kv_mask = default(kv_mask, q_mask)
            mq, mk = bucket(buckets, q_mask), bucket(kv_buckets, kv_mask)
            expand_head_and_merge_into_batch = lambda x: merge_dims(
                0, 1, expand_dim(x.unsqueeze(1), 1, h)
            )
            mq, mk = map(expand_head_and_merge_into_batch, (mq, mk))

            mk_r = batched_index_select(mk, R.abs().argmax(dim=-1))

            if self.n_sortcut > 0:
                mk_r = mk_r[:, 0 : self.n_sortcut].reshape(-1, 1, bsz * self.n_sortcut)
                mk_r = expand_dim(mk_r, 1, buckets)
            else:
                mk_r = mk_r.reshape(bh, buckets, -1)

            mk = torch.cat((mk_r, mk), dim=2) if buckets == kv_buckets else mk_r
            mask = mq[:, :, :, None] * mk[:, :, None, :]
            dots.masked_fill_(~mask, mask_value)
            del mask

        # attention
        dots = dots.softmax(dim=-1)
        dots = self.dropout(dots)

        out = torch.einsum("buij,buje->buie", dots, b_v)
        out = unbucket(out)
        out = out.reshape(b, h, t, d_h)

        out_view = (
            out.transpose(1, 2)
            .contiguous()
            .view(batch_size, -1, self.heads * self.dim_heads)
        )
        output = self.output_linear(out_view)
        return output


def default(x, d):
    if x is None:
        return d if not isfunction(d) else d()
    return x


def max_neg_value(tensor):
    return -torch.finfo(tensor.dtype).max


def batched_index_select(values, indices):
    last_dim = values.shape[-1]
    return values.gather(1, indices[:, :, None].expand(-1, -1, last_dim))


def expand_dim(t, dim, k):
    expand_shape = [-1] * len(t.shape)
    expand_shape[dim] = k
    return t.expand(*expand_shape)


def all_none(*arr):
    return all(el is None for el in arr)


def bucket(buckets, t, dim=1):
    shape = list(t.shape)
    shape[dim : dim + 1] = [buckets, -1]
    return t.reshape(*shape)


def unbucket(t, dim=1):
    shape = list(t.shape)
    shape[dim : dim + 2] = [-1]
    return t.reshape(*shape)


def merge_dims(ind_from, ind_to, tensor):
    shape = list(tensor.shape)
    arr_slice = slice(ind_from, ind_to + 1)
    shape[arr_slice] = [reduce(mul, shape[arr_slice])]
    return tensor.reshape(*shape)


def expand_batch_and_merge_head(b, t):
    shape = list(t.squeeze(0).shape)
    t = expand_dim(t, 0, b)
    shape[0] = shape[0] * b
    return t.reshape(*shape)


def differentiable_topk(x, k, temperature=1.0):
    *_, n, dim = x.shape
    topk_tensors = []

    for i in range(k):
        is_last = i == (k - 1)
        values, indices = (x / temperature).softmax(dim=-1).topk(1, dim=-1)
        topks = torch.zeros_like(x).scatter_(-1, indices, values)
        topk_tensors.append(topks)
        if not is_last:
            x.scatter_(-1, indices, float("-inf"))

    topks = torch.cat(topk_tensors, dim=-1)
    return topks.reshape(*_, k * n, dim)


def reorder_buckets(t, r):
    return torch.einsum("buv,bvtd->butd", r, t)


def sample_gumbel(shape, device, dtype, eps=1e-6):
    u = torch.empty(shape, device=device, dtype=dtype).uniform_(0, 1)
    return -log(-log(u, eps), eps)


def sinkhorn_sorting_operator(r, n_iters=8):
    n = r.shape[1]
    for _ in range(n_iters):
        r = r - torch.logsumexp(r, dim=2, keepdim=True)
        r = r - torch.logsumexp(r, dim=1, keepdim=True)
    return torch.exp(r)


def gumbel_sinkhorn(r, n_iters=8, temperature=0.7):
    r = log(r)
    gumbel = sample_gumbel(r.shape, r.device, r.dtype)
    r = (r + gumbel) / temperature
    return sinkhorn_sorting_operator(r, n_iters)


def log(t, eps=1e-6):
    return torch.log(t + eps)
