import torch
import math
from typing import Optional, Union
from contextlib import nullcontext

from src.components.attention.utils.AttentionMask import AttentionMask

# Taken from library: https://github.com/facebookresearch/xformers/blob/main/xformers/components/attention/core.py


def scaled_query_key_softmax(
    q: torch.Tensor,
    k: torch.Tensor,
    attention_mask: Optional[Union[AttentionMask, torch.Tensor]],
) -> torch.Tensor:

    q = q / math.sqrt(k.size(-1))

    # Matmul with mask
    if attention_mask is not None and isinstance(attention_mask, AttentionMask):
        # Additive mask
        mask: Optional[torch.Tensor] = attention_mask.values
    else:
        mask = attention_mask

    attention = _matmul_with_mask(q, k.transpose(-2, -1), mask)

    # Softmax to get the attention probabilities
    attention = torch.softmax(attention, dim=attention.ndim - 1)

    return attention


def scaled_dot_product_attention(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    attention_mask: Optional[Union[AttentionMask, torch.Tensor]],
    dropout: Optional[torch.nn.Module] = None,
) -> torch.Tensor:
    autocast_disabled = attention_mask is not None and attention_mask.is_sparse

    with torch.cuda.amp.autocast(enabled=False) if autocast_disabled else nullcontext():
        if autocast_disabled:
            q, k, v = q.float(), k.float(), v.float()

        attention = scaled_query_key_softmax(q, k, attention_mask=attention_mask)

        #  Optional dropout, could be part of the masking in the future
        if dropout:
            attention = dropout(attention, dropout)

        # Get to the predicted values, for all heads
        # y = attention @ v  # (N, S, S) x (N, S, hs) -> (N, S, hs)
        y = attention @ v
    return y


def _matmul_with_mask(
    a: torch.Tensor, b: torch.Tensor, mask: Optional[Union[torch.Tensor]]
) -> torch.Tensor:
    if mask is None:
        return a @ b

    attention = a @ b
    if mask.dtype == torch.bool:
        if mask.ndim == 2:
            mask = mask.unsqueeze(0).expand(attention.shape[0], -1, -1)
        # mask is presumed false == ignore
        attention[~mask] = float("-inf")
    else:
        # mask is presumed additive
        attention += mask
    return attention
