import unittest

import torch
from torch import nn

# Inverted Bottleneck (similar to transformer feed forward) = narrow -> wide -> narrow
# From paper: A ConvNet for the 2020s


# NOTE: Dilation >1 turns the module into dilated convolution
class InvertedConvBottleneck(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        forward_expansion: int = 4,
        kernel_size: int = 3,
        dilation: int = 1,
    ):
        super(InvertedConvBottleneck, self).__init__()
        self.embedding_dim = embedding_dim
        self.expanded_dim = embedding_dim * forward_expansion

        # Adjust for even kernel sizes
        if (kernel_size % 2) == 0:
            kernel_size = kernel_size + 1

        # The original paper also includes batch norm after each convolution
        # Each embedding dim is treated as a channel
        self.bottleneck = nn.Sequential(
            # Narrow -> wide
            nn.Conv1d(
                self.embedding_dim,
                self.expanded_dim,
                kernel_size=1,
                stride=1,
            ),
            nn.ReLU(),
            # Wide -> wide (with depth-wise)
            nn.Conv1d(
                self.expanded_dim,
                self.expanded_dim,
                kernel_size=kernel_size,
                stride=1,
                padding=(kernel_size // 2) * dilation,
                groups=embedding_dim,  # This makes it depth-wise
                dilation=dilation,
            ),
            nn.ReLU(),
            # Wide -> narrow
            nn.Conv1d(
                self.expanded_dim,
                self.embedding_dim,
                kernel_size=1,
                stride=1,
            ),
        )

    def forward(self, x):
        x = x.transpose(1, 2)
        out = self.bottleneck(x)
        out = out.transpose(1, 2)
        return out


class TestInvertedConvBottleneck(unittest.TestCase):
    def test(self):
        bottleneck = InvertedConvBottleneck(16, 4, 3, 2)
        input = torch.randn(20, 100, 16)
        output = bottleneck(input)
        self.assertEqual(list(output.shape), [20, 100, 16])
