import torch
from torch import nn

# From paper: Primer: Searching for Efficient Transformers for Language Modeling
# They used it to postprocess keys, queries and values after linear projection


class SpatialDepthWiseConvolution(nn.Module):
    def __init__(self, d_k: int, kernel_size: int = 3):
        super(SpatialDepthWiseConvolution, self).__init__()
        self.kernel_size = kernel_size
        # We set the number of groups to be equal to the number of channels so that it does a separate convolution
        # (with different kernels) for each channel
        # We add padding to both sides and later crop the right most `kernel_size - 1` results
        self.conv = nn.Conv1d(
            in_channels=d_k,
            out_channels=d_k,
            kernel_size=(kernel_size,),
            padding=(kernel_size - 1,),
            groups=d_k,
        )

    def forward(self, x: torch.Tensor):
        batch_size, head_count, sequence_length, d_k = x.shape

        # Permute to (batch_size, heads, d_k, sequence_length)
        x = x.permute(0, 1, 3, 2)
        # Change the shape to (batch_size * heads, d_k, seq_len)
        x = x.reshape(batch_size * head_count, d_k, sequence_length)

        # 1D convolution accepts input of the form (N, channels, sequence)
        x = self.conv(x)
        # Crop the right most `kernel_size - 1` results since we padded both sides
        x = x[:, :, : -(self.kernel_size - 1)]
        # Reshape to (batch_size, heads, d_k, seq_len)
        x = x.view(batch_size, head_count, d_k, sequence_length)
        # Permute to (seq_len, batch_size, heads, d_k)
        x = x.permute(0, 1, 3, 2)
        return x
