import copy
import glob
import os
import random
from enum import Enum
from typing import List, Optional

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from statsmodels.tsa.stattools import adfuller
from torch.utils.data import Dataset

from src.dataset.normalization.Normalizer import Normalizer
from src.dataset.utils.path import get_script_folder
from src.utils.seed import temp_seed


class LibraCategory(str, Enum):
    ECONOMICS = "economics"
    FINANCE = "finance"
    HUMAN = "human"
    NATURE = "nature"

    @staticmethod
    def all_categories() -> List[Enum]:
        return [category for category in LibraCategory]

    @staticmethod
    def from_str(string: str):
        return LibraCategory(string)


def load_category_csvs(path, category, max_datasets) -> List[np.ndarray]:
    if isinstance(category, LibraCategory):
        category = category.value

    glob_path = f"{path}/{category}/*"
    csv_paths = glob.glob(glob_path)
    csv_paths.sort()

    data_frames = [
        pd.read_csv(path, header=None)
        for idx, path in enumerate(csv_paths)
        if idx < max_datasets
    ]

    np_arrays = [frame.to_numpy().flatten() for frame in data_frames]
    return np_arrays


def randomly_sample_datasets(
    categories: List[LibraCategory],
    sampled_datasets_per_category: int,
    x_length: int,
    y_length: int,
    min_training_examples: int = 100,
    seed: int = 1,
) -> list:
    for idx, category in enumerate(categories):
        if isinstance(category, LibraCategory):
            categories[idx] = category.value

    # Load datasets
    datasets_per_category = []
    dataset_path = os.path.join(get_script_folder(__file__), "../../libra")
    for category in categories:
        frames_list = load_category_csvs(dataset_path, category, 100)

        filtered_datasets = []
        for idx, dataset in enumerate(frames_list):
            dataset_length = len(dataset)
            if dataset_length >= (x_length + y_length) + min_training_examples - 1:
                filtered_datasets.append((idx, dataset))

        datasets_per_category.append((category, filtered_datasets))

    # Randomly sample datasets
    with temp_seed(seed):
        datasets_per_category_filtered = []
        for category, datasets in datasets_per_category:
            sampled_datasets_per_category = min(
                sampled_datasets_per_category, len(datasets)
            )
            chosen_datasets = random.sample(datasets, sampled_datasets_per_category)
            datasets_per_category_filtered.append((category, chosen_datasets))

    return datasets_per_category_filtered


def convert_sampled_datasets_to_dataset_filter(datasets_per_category: list) -> list:
    dataset_filter = []
    for _, datasets in datasets_per_category:
        indices = [d[0] for d in datasets]
        dataset_filter.append(indices)
    return dataset_filter


def visualize_datasets(
    category: LibraCategory,
    x_length=400,
    y_length=50,
    max_included_datasets=100,
    min_training_examples_available=100,
    max_values_per_dataset: Optional[int] = None,
    path: str = "../../libra",
):
    loaded_datasets = load_category_csvs(path, category, max_included_datasets)

    dataset_list = []
    for idx, dataset in enumerate(loaded_datasets):
        dataset_length = len(dataset)
        if (
            dataset_length
            >= (x_length + y_length) + min_training_examples_available - 1
        ):
            dataset_list.append((idx, dataset))

    if max_values_per_dataset is not None:
        for idx in range(len(dataset_list)):
            dataset_index = dataset_list[idx][0]
            cropped_dataset = dataset_list[idx][1][:max_values_per_dataset]
            dataset_list[idx] = (dataset_index, cropped_dataset)

    for idx, dataset in dataset_list:
        x = np.arange(0, len(dataset))
        plt.plot(x, dataset)
        plt.title(f"category: {category.value}, dataset index: {idx}", loc="center")
        plt.show()

    total_number_of_examples = [
        len(d) - (x_length + y_length) + 1 for _, d in dataset_list
    ]
    return len(dataset_list), sum(total_number_of_examples)


class LibraDataset(Dataset):
    def __init__(
        self,
        categories: List[LibraCategory],
        x_length=200,
        y_length=50,
        dataset_filter: List[List[int]] = None,
        max_included_datasets=100,
        max_values_per_dataset=5000,
        min_training_examples_available=100,
        normalizer=Normalizer.ZERO_ONE,
        normalization_factor=1.0,
    ):
        self.categories = categories
        self.num_included_datasets = (
            max_included_datasets  # There is 100 datasets per category available
        )
        self.min_training_examples_available = min_training_examples_available  # If less than x training examples can be built from a dataset, exclude it
        self.x_length = x_length
        self.y_length = y_length

        if dataset_filter is not None and len(dataset_filter) != len(categories):
            raise ValueError(
                "Dataset filter list needs to have the same length as categories list"
            )

        self.data_frames = []
        dataset_path = os.path.join(get_script_folder(__file__), "../../libra")
        for idx, category in enumerate(self.categories):
            frames_list = load_category_csvs(
                dataset_path, category, max_included_datasets
            )

            # Only include the datasets at specified indices
            if dataset_filter is not None:
                indices = dataset_filter[idx]
                frames_list = [f for i, f in enumerate(frames_list) if i in indices]

            # Crop to max values per dataset
            if max_values_per_dataset is not None:
                frames_list = [f[:max_values_per_dataset] for f in frames_list]

            # Normalize values
            if normalizer != Normalizer.NONE:
                for i in range(len(frames_list)):
                    norm = normalizer.module(frames_list[i])
                    frames_list[i] = (
                        norm.normalize(frames_list[i]) * normalization_factor
                    )

            self.data_frames.append(frames_list)

        self.index_lookup_table = self.build_index_lookup_table()
        self.total_length = len(self.index_lookup_table)

    def raw_series(self, idx=None):
        if idx is None:
            return copy.deepcopy(self.data_frames)
        return copy.deepcopy(self.data_frames[idx])

    def build_index_lookup_table(self):
        lookup_table = []
        for c_idx, _ in enumerate(self.categories):
            for d_idx, dataset in enumerate(self.data_frames[c_idx]):
                dataset_length = len(dataset)
                if (
                    dataset_length
                    >= (self.x_length + self.y_length)
                    + self.min_training_examples_available
                    - 1
                ):
                    # (category_index, dataset_index, x_start_index, x_end_index, y_start_index, y_end_index)
                    for idx in range(
                        0, dataset_length - (self.x_length + self.y_length) + 1
                    ):
                        lookup_table.append(
                            (
                                c_idx,
                                d_idx,
                                idx,
                                idx + self.x_length,
                                idx + self.x_length,
                                idx + self.x_length + self.y_length,
                            )
                        )
        return lookup_table

    def __len__(self):
        return self.total_length

    def __getitem__(self, idx):
        (
            category_idx,
            dataset_idx,
            x_start_idx,
            x_end_idx,
            y_start_idx,
            y_end_idx,
        ) = self.index_lookup_table[idx]
        dataset = self.data_frames[category_idx][dataset_idx]
        x = dataset[x_start_idx:x_end_idx]
        y = dataset[y_start_idx:y_end_idx]
        return {"x": x, "y": y}

    # Augmented Dickey-Fuller test
    # https://machinelearningmastery.com/time-series-data-stationary-python/
    def is_stationary(self) -> dict:
        ratios_per_category = {}
        for idx, series_list in enumerate(self.data_frames):
            category = self.categories[idx]
            series_count = len(series_list)

            stationary_count = 0
            non_stationary_count = 0
            for series in series_list:
                result = adfuller(series)
                p_value = result[1]

                if p_value < 0.05:
                    stationary_count += 1
                else:
                    non_stationary_count += 1

            non_stationary_ratio = non_stationary_count / series_count
            ratios_per_category[category] = (
                non_stationary_ratio < 0.05,
                non_stationary_ratio,
            )

        return ratios_per_category

    def are_examples_stationary(self) -> (bool, float):
        dataset_length = len(self)
        pairs = [self[i] for i in range(dataset_length)]

        stationary_count = 0
        non_stationary_count = 0
        for idx, p in enumerate(pairs):
            source = p["x"]
            target = p["y"]

            example = np.concatenate((source, target))
            result = adfuller(example)
            p_value = result[1]
            if p_value < 0.05:
                stationary_count += 1
            else:
                non_stationary_count += 1

        non_stationary_ratio = non_stationary_count / dataset_length
        return non_stationary_ratio < 0.05, non_stationary_ratio

    def show_example(self, idx: int):
        item = self[idx]
        x1 = np.arange(0, self.x_length)
        x2 = np.arange(self.x_length, self.x_length + self.y_length)
        plt.plot(x1, item["x"])
        plt.plot(x2, item["y"])
        plt.show()


if __name__ == "__main__":
    all_categories = LibraCategory.all_categories()
    dataset = LibraDataset(
        all_categories,
        normalizer=Normalizer.ZERO_ONE,
        max_values_per_dataset=2000,
        min_training_examples_available=1000,
    )
    dataset_length = len(dataset)

    items = []
    for idx in range(0, dataset_length):
        item = dataset[idx]
        items.append(item)

    print(f"Dataset length: {dataset_length}")
    print(f"Generated examples: {len(items)}\n")

    print(f"Dataset random sampling")
    datasets = randomly_sample_datasets(
        all_categories, 10, 400, 50, min_training_examples=100
    )
    for category, sets in datasets:
        print(f"> {category}: {len(sets)} datasets")

    print(f"Filtered datasets")
    datasets_filter = convert_sampled_datasets_to_dataset_filter(datasets)
    for indices in datasets_filter:
        print(f"> {indices}")

    print(f"Filtered dataset length:")
    dataset2 = LibraDataset(
        all_categories,
        dataset_filter=datasets_filter,
        min_training_examples_available=5000,
        normalizer=Normalizer.ZERO_ONE,
    )
    print(f"Dataset length: {len(dataset2)}")

    filtered_datasets, total_number_of_examples = visualize_datasets(
        LibraCategory.NATURE,
        min_training_examples_available=1000,
        max_values_per_dataset=2000,
    )
    print(f"Shown datasets: {filtered_datasets}")
    print(f"Total number of examples: {total_number_of_examples}\n")
