import copy
from enum import Enum

import numpy as np
from matplotlib import pyplot as plt
from statsmodels.tsa.stattools import adfuller
from torch.utils.data import Dataset

from src.dataset.normalization.Normalizer import Normalizer
from src.dataset.utils.signal import sine_wave, noise
from src.dataset.utils.synthetic import (
    sine_with_sine_periodicity_and_upward_trend_series,
)


class SeriesType(Enum):
    SIMPLE = "simple"
    COMPOSED = "composed"
    COMPOSED_WITH_NOISE = "composed_with_noise"


class SyntheticDataset(Dataset):
    def __init__(
        self,
        series_type,
        length=5000,
        sampling_rate=1,
        x_length=200,
        y_length=50,
        normalizer=Normalizer.ZERO_ONE,
        normalization_factor=1.0,
    ):
        self.series_type = series_type
        self.length = length
        self.sampling_rate = sampling_rate

        self.x_length = x_length
        self.y_length = y_length

        # Params of the series
        frequency1 = 0.03
        amplitude1 = 1.0
        frequency21 = 0.002
        amplitude21 = 1.0
        frequency22 = 0.03
        amplitude22 = 0.5
        trend_slope = 0.0
        noise_mean = 0
        noise_std = 0.1

        if self.series_type == SeriesType.SIMPLE:
            self.series = sine_wave(length, sampling_rate, frequency1, amplitude1)[-1]
        elif self.series_type == SeriesType.COMPOSED:
            self.series = sine_with_sine_periodicity_and_upward_trend_series(
                length,
                sampling_rate,
                frequency21,
                amplitude21,
                frequency22,
                amplitude22,
                trend_slope,
            )[-1]
        elif self.series_type == SeriesType.COMPOSED_WITH_NOISE:
            series1 = sine_with_sine_periodicity_and_upward_trend_series(
                length,
                sampling_rate,
                frequency21,
                amplitude21,
                frequency22,
                amplitude22,
                trend_slope,
            )[-1]
            series2 = noise(length, sampling_rate, mean=noise_mean, std=noise_std)[-1]
            self.series = series1 + series2
        else:
            raise NotImplementedError()

        # Normalize values
        if normalizer != Normalizer.NONE:
            norm = normalizer.module(self.series)
            self.series = norm.normalize(self.series) * normalization_factor

    def raw_series(self):
        return copy.deepcopy(self.series)

    def series_type(self):
        return self.series_type

    # Augmented Dickey-Fuller test
    # https://machinelearningmastery.com/time-series-data-stationary-python/
    def is_stationary(self) -> (bool, float):
        result = adfuller(self.series)
        p_value = result[1]
        return p_value < 0.05, p_value

    def are_examples_stationary(self) -> (bool, float):
        dataset_length = len(self)
        pairs = [self[i] for i in range(dataset_length)]

        stationary_count = 0
        non_stationary_count = 0
        for p in pairs:
            source = p["x"]
            target = p["y"]

            example = np.concatenate((source, target))
            result = adfuller(example)
            p_value = result[1]
            if p_value < 0.05:
                stationary_count += 1
            else:
                non_stationary_count += 1

        non_stationary_ratio = non_stationary_count / dataset_length
        return non_stationary_ratio < 0.05, non_stationary_ratio

    def __len__(self):
        return (self.length * self.sampling_rate) - (self.x_length + self.y_length)

    def __getitem__(self, idx):
        x = self.series[idx : idx + self.x_length]
        y = self.series[idx + self.x_length : idx + self.x_length + self.y_length]
        return {"x": x, "y": y}

    def show(self, idx=None, max_samples=None):
        if idx is not None:
            item = self[idx]
            x1 = np.arange(0, self.x_length)
            x2 = np.arange(self.x_length, self.x_length + self.y_length)
            plt.plot(x1, item["x"])
            plt.plot(x2, item["y"])
            plt.show()
        else:
            series = self.series
            if max_samples is not None:
                series = series[:max_samples]

            x = np.arange(0, len(series))
            plt.plot(x, series)
            plt.show()


if __name__ == "__main__":
    dataset1 = SyntheticDataset(SeriesType.SIMPLE)
    dataset2 = SyntheticDataset(SeriesType.COMPOSED)
    dataset3 = SyntheticDataset(SeriesType.COMPOSED_WITH_NOISE)
    print(f"Dataset lengths: {len(dataset1)}, {len(dataset2)}, {len(dataset3)}")

    dataset1.show(max_samples=300)
    dataset2.show(max_samples=1000)
    dataset3.show(max_samples=1000)

    dataset1.show(0)
    dataset2.show(0)
    dataset3.show(0)
