import importlib
import inspect

from src.evolutionary.genotype.GenotypeCell import GenotypeModuleBranch
from src.evolutionary.genotype.Genotype import Genotype
from src.evolutionary.genotype.GenotypeModule import GenotypeModule
from src.evolutionary.phenotype.Phenotype import (
    Phenotype,
    PhenotypeLayer,
    PhenotypeCell,
)
from src.evolutionary.phenotype.PhenotypeCell import PhenotypeCellBranch


def genotype_to_phenotype(
    genotype: Genotype, input_dim: int, output_dim: int, device
) -> Phenotype:
    embedding_dim = genotype.embedding_dim()
    patch_size = genotype.patch_size()
    dropout = genotype.dropout()

    # IMPORTANT: For modules we need the input_dim after patching!!!
    input_dim_after_splitting_into_patches = input_dim // patch_size
    shared_params = {
        "input_dim": input_dim_after_splitting_into_patches,
        "embedding_dim": embedding_dim,
        "dropout": dropout,
    }

    use_memory = genotype.use_memory()
    use_full_regression = genotype.use_full_regression()
    phenotype = Phenotype(
        embedding_dim,
        input_dim,
        output_dim,
        patch_size,
        use_memory,
        use_full_regression,
    )

    # Preprocess modules
    for module in genotype.preprocess_modules():
        phenotype_module = create_phenotype_module(
            module, shared_params=shared_params
        ).to(device=device)
        phenotype.add_preprocess_module(phenotype_module)

    # Postprocess modules
    for module in genotype.postprocess_modules():
        postprocess_module = create_phenotype_module(
            module, shared_params=shared_params
        ).to(device=device)
        phenotype.add_postprocess_module(postprocess_module)

    # Layers
    for layer_idx in genotype.layer_indices():
        layer = genotype.get_layer(layer_idx)

        phenotype_layer = PhenotypeLayer()
        phenotype.add_layer(phenotype_layer)

        # Cells
        genotype_cells = layer.cells(without_first=True)
        for genotype_cell in genotype_cells:
            phenotype_cell = PhenotypeCell()
            phenotype_layer.add_cell(genotype_cell[0], genotype_cell[1], phenotype_cell)

            # Left branch modules
            for module in genotype_cell[-1].modules(GenotypeModuleBranch.LEFT):
                phenotype_module = create_phenotype_module(
                    module, shared_params=shared_params
                ).to(device=device)
                phenotype_cell.add_module(PhenotypeCellBranch.LEFT, phenotype_module)

            # Right branch modules
            for module in genotype_cell[-1].modules(GenotypeModuleBranch.RIGHT):
                phenotype_module = create_phenotype_module(
                    module, shared_params=shared_params
                ).to(device=device)
                phenotype_cell.add_module(PhenotypeCellBranch.RIGHT, phenotype_module)

    return phenotype


def create_phenotype_module(
    module: GenotypeModule, shared_params=None, override_params=None
):
    if shared_params is None:
        shared_params = {}

    if override_params is None:
        override_params = {}

    module_path = module.get_module_path()
    module_class = module_path.split(".")[-1]
    ModuleClass = module_class_from_name(module_path, module_class)

    # Inspect constructor parameters to see which ones are needed
    # Then filter shared params appropriately
    module_signature = inspect.signature(ModuleClass.__init__)
    module_parameter_keys = list(dict(module_signature.parameters).keys())
    filtered_shared_params = {
        k: v for k, v in shared_params.items() if k in module_parameter_keys
    }
    filtered_override_params = {
        k: v for k, v in override_params.items() if k in module_parameter_keys
    }

    module_params = module.parameters_as_dict()
    merged_params = {
        **filtered_shared_params,
        **module_params,
        **filtered_override_params,
    }  # Module params have priority i.e.override shared params! Override params override both
    phenotype_module = ModuleClass(**merged_params)
    return phenotype_module


def module_class_from_name(module_path: str, class_name: str):
    new_class = getattr(importlib.import_module(module_path), class_name)
    return new_class
