import copy
import json
import random
from enum import Enum
from typing import List

from src.evolutionary.core.bootstrap import (
    bootstrap_pre_norm_transformer_encoder,
    bootstrap_one_module_per_branch_cell,
    bootstrap_evolved_transformer_cell,
)
from src.evolutionary.core.bootstrap import bootstrap_post_norm_transformer_encoder
from src.evolutionary.genotype.Genotype import Genotype
from src.evolutionary.genotype.GenotypeCell import GenotypeCell
from src.evolutionary.genotype.modules.ModuleRegistry import ModuleRegistry
from src.evolutionary.operators.mutation import (
    duplicate_random_cell,
    swap_two_random_modules,
    add_new_random_module,
    change_random_parameter_genotype,
    change_random_parameter,
    remove_random_module,
    duplicate_random_module,
    add_new_cell,
    change_random_cell_inputs,
    delete_random_cell,
    swap_two_random_cells,
    swap_module_for_new_random_module,
    swap_two_random_layers,
    delete_random_layer,
    duplicate_random_layer,
    add_new_layer,
)


class PopulationInitializationMethod(Enum):
    EMPTY_GENOTYPE = "empty_genotype"
    ONE_MODULE_PER_BRANCH = "one_module_per_branch"
    THREE_MODULES_PER_BRANCH = "three_modules_per_branch"
    FULL_ENCODER = "full_encoder"

    @staticmethod
    def from_str(string: str):
        return PopulationInitializationMethod(string)


class TransformerEncoderType(Enum):
    PRE_NORM = "pre_norm"
    POST_NORM = "post_norm"

    @staticmethod
    def from_str(string: str):
        return TransformerEncoderType(string)


def serialize_population(genotypes: List[Genotype]) -> str:
    genotypes_list = []
    for g in genotypes:
        json_str = g.to_json()
        obj = json.loads(json_str)
        genotypes_list.append(obj)
    serialized = json.dumps(genotypes_list)
    return serialized


def deserialize_population(json_str: str) -> List[Genotype]:
    genotypes_list = []
    deserialized = json.loads(json_str)
    for g in deserialized:
        json_str = json.dumps(g)
        obj = Genotype.from_json(json_str)
        genotypes_list.append(obj)
    return genotypes_list


def initialize_population(
    initialization_method: PopulationInitializationMethod,
    template_genotype: Genotype,
    genotype_count: int,
    cell_count: int,
    module_registry: ModuleRegistry,
    encoder_type: TransformerEncoderType,
    prepend_positional_encoding: bool,
    keep_positional_encoding_in_registry: bool,
):
    if prepend_positional_encoding:
        if keep_positional_encoding_in_registry:
            positional_encoding = module_registry.find_module(
                "src.components.positional_encoding.ape.SinusoidalAbsolutePositionalEncoding"
            )
        else:
            positional_encoding = module_registry.pop_module(
                "src.components.positional_encoding.ape.SinusoidalAbsolutePositionalEncoding"
            )
        if positional_encoding is None:
            raise ValueError(
                "Could not find positional encoding module during initialization"
            )
        template_genotype.add_preprocess_module(positional_encoding)

    if initialization_method == PopulationInitializationMethod.EMPTY_GENOTYPE:
        genotypes = initialize_empty_population(
            template_genotype, genotype_count, cell_count
        )
    elif initialization_method == PopulationInitializationMethod.ONE_MODULE_PER_BRANCH:
        genotypes = initialize_one_module_per_branch_population(
            template_genotype, genotype_count, module_registry, cell_count
        )
    elif (
        initialization_method == PopulationInitializationMethod.THREE_MODULES_PER_BRANCH
    ):
        genotypes = initialize_three_modules_per_branch_population(
            template_genotype, genotype_count, module_registry, cell_count
        )
    elif initialization_method == PopulationInitializationMethod.FULL_ENCODER:
        genotypes = initialize_transformer_encoder_population(
            template_genotype, genotype_count, module_registry, encoder_type
        )
    else:
        raise ValueError("Unknown population initialization method")
    return genotypes


def initialize_empty_population(
    template_genotype: Genotype, genotype_count: int, cell_count: int = 2
):
    genotypes = []
    for _ in range(genotype_count):
        genotype = copy.deepcopy(template_genotype)
        layer = genotype.get_layer(0)
        for i in range(cell_count):
            cell = GenotypeCell()
            layer.add_cell(i, i, cell)
        genotypes.append(genotype)
    return genotypes


def initialize_one_module_per_branch_population(
    template_genotype: Genotype,
    genotype_count: int,
    module_registry: ModuleRegistry,
    cell_count: int = 6,
):
    genotypes = []
    for _ in range(genotype_count):
        genotype = copy.deepcopy(template_genotype)
        layer = genotype.get_layer(0)
        for i in range(cell_count):
            cell = bootstrap_one_module_per_branch_cell(module_registry)
            layer.add_cell(i, i, cell)
        genotypes.append(genotype)
    return genotypes


def initialize_three_modules_per_branch_population(
    template_genotype: Genotype,
    genotype_count: int,
    module_registry: ModuleRegistry,
    cell_count: int = 2,
):
    genotypes = []
    for _ in range(genotype_count):
        genotype = copy.deepcopy(template_genotype)
        layer = genotype.get_layer(0)
        for i in range(cell_count):
            cell = bootstrap_evolved_transformer_cell(module_registry)
            layer.add_cell(i, i, cell)
        genotypes.append(genotype)
    return genotypes


def initialize_transformer_encoder_population(
    template_genotype: Genotype,
    genotype_count: int,
    module_registry: ModuleRegistry,
    encoder_type: TransformerEncoderType,
):
    if template_genotype is None:
        template_genotype = Genotype()

    genotypes = []
    for _ in range(genotype_count):
        cloned_template = copy.deepcopy(template_genotype)

        # Pre norm seems like a more plausible solution
        if encoder_type == TransformerEncoderType.PRE_NORM:
            genotype = bootstrap_pre_norm_transformer_encoder(
                cloned_template, module_registry
            )
        elif encoder_type == TransformerEncoderType.POST_NORM:
            genotype = bootstrap_post_norm_transformer_encoder(
                cloned_template, module_registry
            )
        else:
            raise ValueError("Unknown transformer encoder type")

        genotypes.append(genotype)
    return genotypes


def randomize_population(
    genotypes: List[Genotype],
    available_modules: list,
    independent_layer_evolution: bool,
    rand_layers_iterations: int,
    rand_cells_iterations: int,
    rand_modules_iterations: int,
    rand_parameters_iterations: int,
):
    for g in genotypes:
        if independent_layer_evolution:
            randomize_layers(g, iterations=rand_layers_iterations)
        randomize_cells(g, iterations=rand_cells_iterations)
        randomize_modules(
            g, iterations=rand_modules_iterations, available_modules=available_modules
        )
        randomize_parameters(g, iterations=rand_parameters_iterations)


def randomize_layers(genotype: Genotype, iterations: int):
    for _ in range(iterations):
        choice = random.random()
        if choice < 0.25:
            add_new_layer(genotype)
        elif 0.25 < choice < 0.5:
            duplicate_random_layer(genotype)
        elif 0.5 < choice < 0.75:
            swap_two_random_layers(genotype)
        else:
            delete_random_layer(genotype)
    genotype.mark_dirty()


def randomize_cells(genotype: Genotype, iterations: int):
    for _ in range(iterations):
        choice = random.random()
        if choice < 0.2:
            add_new_cell(genotype)
        elif 0.2 < choice < 0.4:
            duplicate_random_cell(genotype)
        elif 0.4 < choice < 0.6:
            swap_two_random_cells(genotype)
        elif 0.6 < choice < 0.8:
            change_random_cell_inputs(genotype)
        else:
            delete_random_cell(genotype)
    genotype.mark_dirty()


def randomize_modules(genotype: Genotype, available_modules: list, iterations: int):
    for _ in range(iterations):
        choice = random.random()
        if choice < 0.2:
            add_new_random_module(genotype, available_modules)
        elif 0.2 < choice < 0.4:
            duplicate_random_module(genotype)
        elif 0.4 < choice < 0.6:
            swap_two_random_modules(genotype)
        elif 0.6 < choice < 0.8:
            respect_category = random.random() < 0.5
            swap_module_for_new_random_module(
                genotype, available_modules, respect_category
            )
        else:
            remove_random_module(genotype)
    genotype.mark_dirty()


def randomize_parameters(genotype: Genotype, iterations: int):
    for _ in range(iterations):
        if random.random() < 0.5:
            change_random_parameter_genotype(genotype)
        else:
            change_random_parameter(genotype)
    genotype.mark_dirty()
