import json
import re
from typing import List, Dict, Optional

from src.evolutionary.genotype.parameters.IParameter import IParameter
from src.evolutionary.genotype.parameters.ParameterType import ParameterType


class GenotypeModule:
    def __init__(
        self,
        module_path: str,
        parameters: Dict[str, IParameter] = None,
        category: Optional[str] = None,
        tags: Optional[List[str]] = None,
    ):
        if parameters is None:
            parameters = {}
        if tags is None:
            tags = []

        self.module_path = module_path
        self.parameters = parameters
        self.category = category
        self.tags = tags

    def register_parameter(self, name: str, parameter: IParameter):
        self.parameters[name] = parameter
        return self

    def get_parameter(self, key: str) -> IParameter:
        if key in self.parameters:
            return self.parameters[key]
        else:
            raise IndexError("Layer index is out of bounds")

    def parameter_count(self) -> int:
        return len(self.parameters)

    def get_parameters(self) -> List[IParameter]:
        key_values = list(self.parameters.items())
        key_values_sorted = sorted(key_values, key=lambda x: x[0])
        params = [p for _, p in key_values_sorted]
        return params

    def parameter_keys(self) -> List[str]:
        keys = list(self.parameters.keys())
        keys.sort()
        return keys

    def parameters_as_dict(self) -> Dict[str, any]:
        params = {k: v.get_value() for k, v in self.parameters.items()}
        return params

    def get_module_path(self, absolute=True) -> str:
        if absolute:
            return self.module_path
        else:
            return self.module_path.split(".")[-1]

    # Categories
    def get_category(self) -> str:
        return self.category

    def set_category(self, category: Optional[str]):
        self.category = category

    def is_attention(self) -> bool:
        return self.category == "attention"

    def is_normalization(self) -> bool:
        return self.category == "normalization"

    # Tags
    def add_tag(self, tag: str):
        return self.tags.append(tag)

    def remove_tag(self, tag: str):
        self.tags.remove(tag)

    def get_tags(self) -> List[str]:
        return self.tags

    def get_module_label(self, abbr_length=3, include_category=False) -> str:
        module_name = self.module_path.split(".")[-1]
        split_on_upper_case = re.findall("[A-Z][^A-Z]*", module_name)

        def remove_vowels(string: str):
            vowels = ("a", "e", "i", "o", "u", "y")
            return "".join([c for c in string if c not in vowels])

        abbrs = [split[:abbr_length] for split in split_on_upper_case]
        rest = [remove_vowels(split[abbr_length:]) for split in split_on_upper_case]
        module_label = "".join([a + r for a, r in zip(abbrs, rest)])

        if include_category:
            module_label += f"\n({self.category.replace('_', ' ')})"

        return module_label

    def to_dict(self) -> dict:
        data = dict(self.__dict__)
        data["parameters"] = {k: v.to_dict() for k, v in data["parameters"].items()}
        return data

    @staticmethod
    def from_dict(data: dict):
        module_path = data["module_path"]
        category = data["category"]
        tags = data["tags"]

        parameters = {}
        for key, param in data["parameters"].items():
            param_type = param.pop("type")
            parameter = ParameterType.from_str(param_type).instance(**param)
            parameters[key] = parameter

        return GenotypeModule(module_path, parameters, category, tags)

    def save_as_json(self, filename: str):
        with open(filename, "w") as outfile:
            json_str = json.dumps(self.to_dict(), indent=4)
            outfile.write(json_str)

    @staticmethod
    def load_from_json(filename: str):
        with open(filename, "r") as infile:
            json_string = infile.read()
            dict_data = json.loads(json_string)
            return GenotypeModule.from_dict(dict_data)
