import unittest
import torch

from src.evolutionary.core.conversion import create_phenotype_module
from src.evolutionary.genotype.GenotypeModule import GenotypeModule
from src.evolutionary.genotype.modules.ModuleCategory import ModuleCategory
from src.evolutionary.genotype.parameters.RangeParameter import RangeParameter
from src.evolutionary.genotype.parameters.SetParameter import SetParameter


modules = [
    GenotypeModule(
        "src.components.positional_encoding.ape.SinusoidalAbsolutePositionalEncoding",
        {
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.POSITIONAL_ENCODING.value,
    ),
    # The modules are set to generic to make them swappable when respect_category is True
    GenotypeModule(
        "src.components.attention.ScaledDotProductMultiHeadAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
        },
        ModuleCategory.GENERIC.value,
    ),
    GenotypeModule(
        "src.components.transformer.FeedForward",
        {
            "forward_expansion": RangeParameter(1, (1, 8)),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.GENERIC.value,
    ),
    GenotypeModule(
        "src.components.feed_forward.SingleFeedForward",
        {},
        ModuleCategory.GENERIC.value,
    ),
    GenotypeModule(
        "src.components.normalization.LayerNormalization",
        {},
        ModuleCategory.NORMALIZATION.value,
    ),
    GenotypeModule(
        "src.components.activation.ReLU",
        {},
        ModuleCategory.ACTIVATION.value,
    ),
    # Passthrough modules for each category i.e. no op
    GenotypeModule(
        "src.components.transformer.Passthrough",
        {},
        ModuleCategory.GENERIC.value,
    ),
    GenotypeModule(
        "src.components.transformer.Passthrough",
        {},
        ModuleCategory.NORMALIZATION.value,
    ),
    GenotypeModule(
        "src.components.transformer.Passthrough",
        {},
        ModuleCategory.ACTIVATION.value,
    ),
]


class TestExperimentModules(unittest.TestCase):
    def test_instantiation_of_all_modules(self):
        shared_params = {"embedding_dim": 16, "dropout": 0.1}
        phenotype_modules = []
        for module in modules:
            phenotype_module = create_phenotype_module(
                module, shared_params=shared_params
            )
            phenotype_modules.append(phenotype_module)
        self.assertEqual(len(phenotype_modules), len(modules))

    def test_all_parameters(self):
        batch_size = 16
        dropout = 0.1

        # The shortest possible sequence length is 8 i.e. patch_size = 50 / input_dim = 400
        # Available patch sizes: [1, 5, 10, 25, 50]
        sequence_lengths = [8, 20, 40, 200]
        embedding_dims = [16, 32, 64]

        for ed in embedding_dims:
            for sl in sequence_lengths:
                for module in modules:
                    input_tensor = torch.rand(batch_size, sl, ed)
                    shared_params = {
                        "embedding_dim": ed,
                        "input_dim": sl,
                        "dropout": dropout,
                    }

                    for key, param in module.parameters.items():
                        allowed_values = param.get_allowed_values()
                        for value in allowed_values:
                            override_params = {key: value}
                            try:
                                phenotype_module = create_phenotype_module(
                                    module,
                                    shared_params=shared_params,
                                    override_params=override_params,
                                )
                                output_tensor = phenotype_module(input_tensor)
                            except Exception as e:
                                print(
                                    f"embedding_dim: {ed}, input_dim: {sl}, module: {module.module_path}"
                                )
                                print(f"param: {key}, value: {value}")
                                raise e

                            self.assertEqual(output_tensor.shape, (batch_size, sl, ed))
