import unittest
from typing import Tuple, TypeVar, List

from src.evolutionary.genotype.parameters.IParameter import IParameter

T = TypeVar("T", int, float)


class NeighbourhoodParameter(IParameter):
    def __init__(
        self, initial_value: T, step: T = 0.1, limits: Tuple[T, T] = (0.0, 1.0)
    ):
        self.value: T = initial_value
        self.limits: Tuple[T, T] = limits
        self.step: T = step
        self.type = "neighbourhood"

    def get_value(self):
        return self.value

    def set_value(self, value: T):
        if value not in self.get_allowed_values():
            raise ValueError("Value not allowed")
        self.value = value

    def get_allowed_values(self) -> List[T]:
        allowed_values = [
            self.value - self.step,
            self.value,
            self.value + self.step,
        ]
        allowed_values_clamped = [
            v for v in allowed_values if self.limits[0] <= v <= self.limits[1]
        ]
        return allowed_values_clamped

    # Serialize/deserialize
    def to_dict(self) -> dict:
        data = {
            "value": self.value,
            "step": self.step,
            "limits": self.limits,
            "type": self.type,
        }
        return data

    @staticmethod
    def from_dict(data: dict):
        value = data["value"]
        step = data["step"]
        limits = data["limits"]
        return NeighbourhoodParameter(value, step, limits)


class TestNeighbourhoodParameter(unittest.TestCase):
    def test(self):
        parameter = NeighbourhoodParameter(0, 0.1)
        allowed_values = parameter.get_allowed_values()
        self.assertEqual(allowed_values, [0.0, 0.1])
        self.assertEqual(parameter.get_value(), 0)
        parameter.set_value(0.1)
        self.assertEqual(parameter.get_value(), 0.1)
