import click

from src.evolutionary.genotype.modules.ModuleRegistry import ModuleRegistry
from src.evolutionary.genotype.modules.registries.all_activation import (
    modules as all_activation,
)
from src.evolutionary.genotype.modules.registries.all_attentions import (
    modules as all_attentions,
)
from src.evolutionary.genotype.modules.registries.all_convolution import (
    modules as all_convolution,
)
from src.evolutionary.genotype.modules.registries.all_dropout import (
    modules as all_dropout,
)
from src.evolutionary.genotype.modules.registries.all_feed_forward import (
    modules as all_feed_forward,
)
from src.evolutionary.genotype.modules.registries.all_normalization import (
    modules as all_normalization,
)
from src.evolutionary.genotype.modules.registries.all_positional_encodings import (
    modules as all_positional_encodings,
)
from src.evolutionary.main.config import (
    DatasetConfig,
    EvolutionConfig,
    EvaluationConfig,
    GenotypeConfig,
    GlobalConfig,
)
from src.evolutionary.main.main import main
from src.utils.logger import logger

log = logger()


@click.command()
@click.option(
    "-dc",
    "--dataset-config",
    type=click.Path(),
    help=f"JSON config file: {str(DatasetConfig.fields())}",
)
@click.option(
    "-evolc",
    "--evolution-config",
    type=click.Path(),
    help=f"JSON config file: {str(EvolutionConfig.fields())}",
)
@click.option(
    "-evalc",
    "--evaluation-config",
    type=click.Path(),
    help=f"JSON config file: {str(EvaluationConfig.fields())}",
)
@click.option(
    "-gc",
    "--genotype-config",
    type=click.Path(),
    help=f"JSON config file: {str(GenotypeConfig.fields())}",
)
@click.option(
    "-c",
    "--global-config",
    type=click.Path(),
    help=f"JSON config file: {str(GlobalConfig.fields())}",
)
@click.option(
    "-mr", "--module-registry", type=click.Path(), help=f"Path to module registry"
)
@click.option(
    "-jc",
    "--just-config",
    type=click.Path(),
    help=f"Just creates configs and saves them to json",
)
def nas(
    dataset_config,
    evolution_config,
    evaluation_config,
    genotype_config,
    global_config,
    module_registry,
    just_config,
):
    # USe global config if supplied
    if global_config:
        log.info("%%%%%% Using global config %%%%%%")
        config = GlobalConfig.load_from_json(global_config)
    # If not, use individual configs
    elif (
        dataset_config
        and evolution_config
        and evaluation_config
        and genotype_config
        and module_registry
    ):
        log.info("%%%%%% Using individual configs %%%%%%")
        dataset_config = DatasetConfig.load_from_json(dataset_config)
        evolution_config = EvolutionConfig.load_from_json(evolution_config)
        evaluation_config = EvaluationConfig.load_from_json(evaluation_config)
        genotype_config = GenotypeConfig.load_from_json(genotype_config)
        module_registry = ModuleRegistry.load_from_json(module_registry)
        config_dict = {
            "dataset": dataset_config,
            "evolution": evolution_config,
            "evaluation": evaluation_config,
            "genotype": genotype_config,
            "modules": module_registry.modules(),
        }
        config = GlobalConfig(**config_dict)
    # If only module registry is available, use it and also the default values from config.py
    elif module_registry:
        log.info("%%%%%% Using default config with module registry %%%%%%")
        module_registry = ModuleRegistry.load_from_json(module_registry)
        default_configs_dict = {
            "dataset": DatasetConfig().to_dict(),
            "evolution": EvolutionConfig().to_dict(),
            "evaluation": EvaluationConfig().to_dict(),
            "genotype": GenotypeConfig().to_dict(),
            "modules": module_registry.modules(),
        }
        config = GlobalConfig(**default_configs_dict)
    # If nothing is available, use default values from config.py
    else:
        log.info("%%%%%% Using only default config %%%%%%")
        # Use all available modules from all registries
        all_modules = (
            all_activation
            + all_attentions
            + all_convolution
            + all_dropout
            + all_feed_forward
            + all_normalization
            + all_positional_encodings
        )
        all_modules_as_dicts = [m.to_dict() for m in all_modules]
        default_configs_dict = {
            "dataset": DatasetConfig().to_dict(),
            "evolution": EvolutionConfig().to_dict(),
            "evaluation": EvaluationConfig().to_dict(),
            "genotype": GenotypeConfig().to_dict(),
            "modules": all_modules_as_dicts,
        }
        config = GlobalConfig(**default_configs_dict)

    # Run the evolution or just save the config
    if just_config:
        config.save_as_json(just_config)
    else:
        main(config)


if __name__ == "__main__":
    nas()
