import copy
import random
import unittest
from typing import List

from src.evolutionary.genotype.GenotypeCell import GenotypeCell, GenotypeModuleBranch
from src.evolutionary.genotype.Genotype import Genotype
from src.evolutionary.genotype.GenotypeLayer import GenotypeLayer
from src.evolutionary.genotype.GenotypeModule import GenotypeModule


# Layer mutations
def add_new_layer(genotype: Genotype, layer_template: GenotypeLayer = None):
    layer_count = genotype.layer_count()
    layer_insert_index = random.randint(0, layer_count)

    if layer_template is None:
        new_layer = GenotypeLayer()
    else:
        new_layer = copy.deepcopy(layer_template)

    genotype.add_layer(new_layer, layer_insert_index)


def duplicate_random_layer(genotype: Genotype):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        random_layer = genotype.get_layer(random_layer_index)
        random_layer_clone = copy.deepcopy(random_layer)  # Important!

        layer_count = genotype.layer_count()
        layer_insert_index = random.randint(0, layer_count)
        genotype.add_layer(random_layer_clone, layer_insert_index)


def swap_two_random_layers(genotype: Genotype):
    layer_count = genotype.layer_count()
    if layer_count > 1:
        layer_indices = genotype.layer_indices()

        random_layer1_index = random.choice(layer_indices)
        random_layer1 = genotype.delete_layer(random_layer1_index)

        random_layer2_index = random.choice(layer_indices)
        random_layer2 = genotype.delete_layer(random_layer2_index)

        genotype.add_layer(random_layer1, random_layer2_index)
        genotype.add_layer(random_layer2, random_layer1_index)


def delete_random_layer(genotype: Genotype):
    layer_count = genotype.layer_count()
    if layer_count > 1:  # Prevent removal of all layers!
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        genotype.delete_layer(random_layer_index)


# Cell mutations
def add_new_cell(genotype: Genotype, cell_template: GenotypeCell = None):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        cell_insert_index = random.randint(0, cell_count)

        hidden_state_indices = layer.hidden_state_indices()
        hidden_state_indices_filtered = [
            i for i in hidden_state_indices if i <= cell_insert_index
        ]
        left_index = random.choice(hidden_state_indices_filtered)
        right_index = random.choice(hidden_state_indices_filtered)

        if cell_template is None:
            new_cell = GenotypeCell()
        else:
            new_cell = copy.deepcopy(cell_template)

        layer.add_cell(left_index, right_index, new_cell, cell_insert_index)


def duplicate_random_cell(genotype: Genotype):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        if cell_count > 0:
            cell_indices = layer.cell_indices()
            random_cell_index = random.choice(cell_indices)
            random_cell = layer.get_cell(random_cell_index)
            random_cell_clone = copy.deepcopy(random_cell)  # Important!

            cell_insert_index = random.randint(0, cell_count)

            hidden_state_indices = layer.hidden_state_indices()
            hidden_state_indices_filtered = [
                i for i in hidden_state_indices if i <= cell_insert_index
            ]
            left_index = random.choice(hidden_state_indices_filtered)
            right_index = random.choice(hidden_state_indices_filtered)

            layer.add_cell(
                left_index, right_index, random_cell_clone, cell_insert_index
            )


def change_random_cell_inputs(genotype: Genotype):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        if cell_count > 0:
            cell_indices = layer.cell_indices()
            random_cell_index = random.choice(cell_indices)

            hidden_state_indices = layer.hidden_state_indices()
            hidden_state_indices_filtered = [
                i for i in hidden_state_indices if i < random_cell_index
            ]
            left_index = random.choice(hidden_state_indices_filtered)
            right_index = random.choice(hidden_state_indices_filtered)
            layer.set_cell_hidden_state_indices(
                random_cell_index, left_index, right_index
            )


def swap_two_random_cells(genotype: Genotype):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        if cell_count > 1:
            cell_indices = layer.cell_indices()
            random_cell1_index, random_cell2_index = random.sample(cell_indices, 2)

            random_cell1_hidden_state_indices = layer.get_cell_hidden_state_indices(
                random_cell1_index
            )
            random_cell2_hidden_state_indices = layer.get_cell_hidden_state_indices(
                random_cell2_index
            )

            random_cell1 = layer.get_cell(random_cell1_index)
            random_cell2 = layer.get_cell(random_cell2_index)

            layer.add_cell(
                random_cell1_hidden_state_indices[0],
                random_cell1_hidden_state_indices[1],
                random_cell2,
                random_cell1_index,
            )
            _deleted_cell1 = layer.delete_cell(
                random_cell1_index + 1
            )  # The old cell will get pushed to i + 1

            layer.add_cell(
                random_cell2_hidden_state_indices[0],
                random_cell2_hidden_state_indices[1],
                random_cell1,
                random_cell2_index,
            )
            _deleted_cell2 = layer.delete_cell(
                random_cell2_index + 1
            )  # The old cell will get pushed to i + 1


def delete_random_cell(genotype: Genotype):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        if cell_count > 1:  # Prevent removal of all cells!
            cell_indices = layer.cell_indices()
            random_cell_index = random.choice(cell_indices)

            # Removal of the cell will shift indices for the following cells
            for cell in layer.cells():
                if cell[0] >= random_cell_index:
                    cell[0] -= 1
                if cell[1] >= random_cell_index:
                    cell[1] -= 1
            layer.delete_cell(random_cell_index)


# Module mutations
# Modules on the root
def add_new_random_module_root(
    genotype: Genotype, available_modules: List[GenotypeModule]
):
    random_module = random.choice(available_modules)
    random_module_clone = copy.deepcopy(random_module)  # Important!

    # Choose between pre- and post-process modules
    if random.random() < 0.5:
        module_count = genotype.preprocess_module_count()
        module_insert_index = random.randint(0, module_count)
        genotype.add_preprocess_module(random_module_clone, module_insert_index)
    else:
        module_count = genotype.postprocess_module_count()
        module_insert_index = random.randint(0, module_count)
        genotype.add_postprocess_module(random_module_clone, module_insert_index)


def duplicate_random_module_root(genotype: Genotype):
    # Choose between pre- and post-process modules
    if random.random() < 0.5:
        random_module = random.choice(genotype.preprocess_modules())
        random_module_clone = copy.deepcopy(random_module)  # Important!
    else:
        random_module = random.choice(genotype.postprocess_modules())
        random_module_clone = copy.deepcopy(random_module)  # Important!

    # Choose between pre- and post-process modules
    if random.random() < 0.5:
        module_count = genotype.preprocess_module_count()
        module_insert_index = random.randint(0, module_count)
        genotype.add_preprocess_module(random_module_clone, module_insert_index)
    else:
        module_count = genotype.postprocess_module_count()
        module_insert_index = random.randint(0, module_count)
        genotype.add_postprocess_module(random_module_clone, module_insert_index)


def swap_two_random_modules_root(genotype: Genotype):
    # Choose between pre- and post-process modules
    if random.random() < 0.5:
        random_module_index = random.choice(genotype.preprocess_module_indices())
        random_module = genotype.delete_preprocess_module(random_module_index)
    else:
        random_module_index = random.choice(genotype.postprocess_module_indices())
        random_module = genotype.delete_postprocess_module(random_module_index)

    # Choose between pre- and post-process modules
    if random.random() < 0.5:
        module_count = genotype.preprocess_module_count()
        module_insert_index = random.randint(0, module_count)
        genotype.add_preprocess_module(random_module, module_insert_index)
    else:
        module_count = genotype.postprocess_module_count()
        module_insert_index = random.randint(0, module_count)
        genotype.add_postprocess_module(random_module, module_insert_index)


def swap_module_for_new_random_module_root(
    genotype: Genotype, available_modules: List[GenotypeModule], respect_category=False
):
    # Choose between pre- and post-process modules
    pre_or_post = random.random() < 0.5
    if pre_or_post:
        random_module_index = random.choice(genotype.preprocess_module_indices())
        random_module = genotype.delete_preprocess_module(random_module_index)
    else:
        random_module_index = random.choice(genotype.postprocess_module_indices())
        random_module = genotype.delete_postprocess_module(random_module_index)

    if respect_category:
        random_module_category = random_module.get_category()
        modules = [
            m for m in available_modules if m.get_category() == random_module_category
        ]
    else:
        modules = available_modules

    new_module = random.choice(modules)

    if pre_or_post:
        genotype.add_preprocess_module(new_module, random_module_index)
    else:
        genotype.add_postprocess_module(new_module, random_module_index)


def remove_random_module_root(genotype: Genotype):
    # Choose between pre- and post-process modules
    if random.random() < 0.5:
        module_indices = genotype.preprocess_module_indices()
        module_index = random.choice(module_indices)
        genotype.delete_preprocess_module(module_index)
    else:
        module_indices = genotype.postprocess_module_indices()
        module_index = random.choice(module_indices)
        genotype.delete_postprocess_module(module_index)


# Modules on layers
def add_new_random_module(genotype: Genotype, available_modules: List[GenotypeModule]):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        if cell_count > 0:
            cell_indices = layer.cell_indices()
            random_cell_index = random.choice(cell_indices)
            cell = layer.get_cell(random_cell_index)

            random_module = random.choice(available_modules)
            random_module_clone = copy.deepcopy(random_module)  # Important!
            branch_index = random.uniform(0, 1)

            branch = (
                GenotypeModuleBranch.LEFT
                if branch_index < 0.5
                else GenotypeModuleBranch.RIGHT
            )
            module_count = cell.module_count(branch)
            module_insert_index = random.randint(0, module_count)
            cell.add_module(random_module_clone, branch, module_insert_index)


def duplicate_random_module(genotype: Genotype):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        if cell_count > 0:
            cell_indices = layer.cell_indices()
            random_cell_index = random.choice(cell_indices)
            cell = layer.get_cell(random_cell_index)

            # Select random branch
            branch_index = random.uniform(0, 1)
            branch = (
                GenotypeModuleBranch.LEFT
                if branch_index < 0.5
                else GenotypeModuleBranch.RIGHT
            )
            module_count = cell.module_count(branch)

            # If the selected branch has no modules, select the other one
            if module_count == 0:
                branch = (
                    GenotypeModuleBranch.RIGHT
                    if branch == GenotypeModuleBranch.LEFT
                    else GenotypeModuleBranch.LEFT
                )
                module_count = cell.module_count(branch)

            if module_count > 0:
                random_module_index = random.randint(0, module_count - 1)
                random_module = cell.get_module(branch, random_module_index)
                random_module_clone = copy.deepcopy(random_module)  # Important!

                new_module_branch_index = random.uniform(0, 1)
                new_module_branch = (
                    GenotypeModuleBranch.LEFT
                    if new_module_branch_index < 0.5
                    else GenotypeModuleBranch.RIGHT
                )
                new_module_branch_module_count = cell.module_count(new_module_branch)
                new_module_branch_module_insert_index = random.randint(
                    0, new_module_branch_module_count
                )
                cell.add_module(
                    random_module_clone,
                    new_module_branch,
                    new_module_branch_module_insert_index,
                )


def swap_two_random_modules(genotype: Genotype):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        if cell_count > 0:
            cell_indices = layer.cell_indices()
            random_cell_index = random.choice(cell_indices)
            cell = layer.get_cell(random_cell_index)

            # Select random branch 1
            branch1_index = random.uniform(0, 1)
            branch1 = (
                GenotypeModuleBranch.LEFT
                if branch1_index < 0.5
                else GenotypeModuleBranch.RIGHT
            )
            module_count1 = cell.module_count(branch1)

            # If the selected branch has no modules, select the other one
            if module_count1 == 0:
                branch1 = (
                    GenotypeModuleBranch.RIGHT
                    if branch1 == GenotypeModuleBranch.LEFT
                    else GenotypeModuleBranch.LEFT
                )
                module_count1 = cell.module_count(branch1)

            # Select random branch 2
            branch2_index = random.uniform(0, 1)
            branch2 = (
                GenotypeModuleBranch.LEFT
                if branch2_index < 0.5
                else GenotypeModuleBranch.RIGHT
            )
            module_count2 = cell.module_count(branch2)

            # If the selected branch has no modules, select the other one
            if module_count2 == 0:
                branch2 = (
                    GenotypeModuleBranch.RIGHT
                    if branch2 == GenotypeModuleBranch.LEFT
                    else GenotypeModuleBranch.LEFT
                )
                module_count2 = cell.module_count(branch2)

            if branch1 == branch2 and module_count1 > 1:
                random_module1_index = random.randint(0, module_count1 - 1)
                random_module1 = cell.delete_module(branch1, random_module1_index)

                # One module has already been removed, so -2
                random_module2_index = random.randint(0, module_count1 - 2)
                random_module2 = cell.delete_module(branch1, random_module2_index)

                cell.add_module(random_module1, branch1, random_module2_index)
                cell.add_module(random_module2, branch1, random_module1_index)

            elif branch1 != branch2 and module_count1 > 0 and module_count2 > 0:
                random_module1_index = random.randint(0, module_count1 - 1)
                random_module1 = cell.delete_module(branch1, random_module1_index)

                random_module2_index = random.randint(0, module_count2 - 1)
                random_module2 = cell.delete_module(branch2, random_module2_index)

                cell.add_module(random_module1, branch2, random_module2_index)
                cell.add_module(random_module2, branch1, random_module1_index)


def swap_module_for_new_random_module(
    genotype: Genotype, available_modules: List[GenotypeModule], respect_category=False
):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        if cell_count > 0:
            cell_indices = layer.cell_indices()
            random_cell_index = random.choice(cell_indices)
            cell = layer.get_cell(random_cell_index)

            # Select random branch
            branch_index = random.uniform(0, 1)
            branch = (
                GenotypeModuleBranch.LEFT
                if branch_index < 0.5
                else GenotypeModuleBranch.RIGHT
            )
            cell_module_indices = cell.module_indices(branch)

            # If the selected branch has no modules, select the other one
            if len(cell_module_indices) == 0:
                branch = (
                    GenotypeModuleBranch.RIGHT
                    if branch == GenotypeModuleBranch.LEFT
                    else GenotypeModuleBranch.LEFT
                )
                cell_module_indices = cell.module_indices(branch)

            if len(cell_module_indices) > 0:
                random_module_index = random.choice(cell_module_indices)
                random_module = cell.delete_module(branch, random_module_index)
                random_module_category = random_module.get_category()

                if respect_category:
                    modules = [
                        m
                        for m in available_modules
                        if m.get_category() == random_module_category
                    ]
                else:
                    modules = available_modules

                new_module = random.choice(modules)
                cell.add_module(new_module, branch, random_module_index)


def remove_random_module(genotype: Genotype):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        if cell_count > 0:
            cell_indices = layer.cell_indices()
            random_cell_index = random.choice(cell_indices)
            cell = layer.get_cell(random_cell_index)

            # Select random branch
            branch_index = random.uniform(0, 1)
            branch = (
                GenotypeModuleBranch.LEFT
                if branch_index < 0.5
                else GenotypeModuleBranch.RIGHT
            )
            module_count = cell.module_count(branch)

            # If the selected branch has no modules, select the other one
            if module_count == 0:
                branch = (
                    GenotypeModuleBranch.RIGHT
                    if branch == GenotypeModuleBranch.LEFT
                    else GenotypeModuleBranch.LEFT
                )
                module_count = cell.module_count(branch)

            if module_count > 0:
                module_indices = cell.module_indices(branch)
                random_module_index = random.choice(module_indices)
                cell.delete_module(branch, random_module_index)


# Parameter mutations
# Parameters on genotype level
def change_random_parameter_genotype(genotype: Genotype):
    parameters = genotype.parameters()
    parameter = random.choice(parameters)

    allowed_parameter_values = parameter.get_allowed_values()
    new_parameter_value = random.choice(allowed_parameter_values)
    parameter.set_value(new_parameter_value)


# Parameters on preprocess and postprocess modules
def change_random_parameter_root(genotype: Genotype):
    modules = genotype.preprocess_modules() + genotype.postprocess_modules()
    module = random.choice(modules)

    parameter_count = module.parameter_count()
    if parameter_count > 0:
        parameter_keys = module.parameter_keys()
        random_parameter_key = random.choice(parameter_keys)
        parameter = module.get_parameter(random_parameter_key)

        allowed_parameter_values = parameter.get_allowed_values()
        new_parameter_value = random.choice(allowed_parameter_values)
        parameter.set_value(new_parameter_value)


# Parameters on layers
def change_random_parameter(genotype: Genotype, number_of_inspected_modules=5):
    layer_count = genotype.layer_count()
    if layer_count > 0:
        layer_indices = genotype.layer_indices()
        random_layer_index = random.choice(layer_indices)
        layer = genotype.get_layer(random_layer_index)

        cell_count = layer.cell_count()
        if cell_count > 0:
            cell_indices = layer.cell_indices()
            random_cell_index = random.choice(cell_indices)
            cell = layer.get_cell(random_cell_index)

            # Select random branch
            branch_index = random.uniform(0, 1)
            branch = (
                GenotypeModuleBranch.LEFT
                if branch_index < 0.5
                else GenotypeModuleBranch.RIGHT
            )
            module_count = cell.module_count(branch)

            # If the selected branch has no modules, select the other one
            if module_count == 0:
                branch = (
                    GenotypeModuleBranch.RIGHT
                    if branch == GenotypeModuleBranch.LEFT
                    else GenotypeModuleBranch.LEFT
                )
                module_count = cell.module_count(branch)

            if module_count > 0:
                module_indices = cell.module_indices(branch)

                # Select a random module
                module_index = random.choice(module_indices)
                module = cell.get_module(branch, module_index)
                parameter_count = module.parameter_count()

                # Try several other randomly selected modules and look for evolvable parameters
                if parameter_count == 0:
                    for _ in range(max(number_of_inspected_modules - 1, 0)):
                        module_index_inner = random.choice(module_indices)
                        module_inner = cell.get_module(branch, module_index_inner)
                        parameter_count_inner = module_inner.parameter_count()

                        if parameter_count_inner > 0:
                            parameter_count = parameter_count_inner
                            module = module_inner
                            break

                if parameter_count > 0:
                    parameter_keys = module.parameter_keys()
                    random_parameter_key = random.choice(parameter_keys)
                    parameter = module.get_parameter(random_parameter_key)

                    allowed_parameter_values = parameter.get_allowed_values()
                    new_parameter_value = random.choice(allowed_parameter_values)
                    parameter.set_value(new_parameter_value)


class TestMutations(unittest.TestCase):
    def test_layer_mutation(self):
        genotype = Genotype()
        layer1 = GenotypeLayer()
        layer2 = GenotypeLayer()
        genotype.add_layer(layer1)
        genotype.add_layer(layer2)
        self.assertEqual(genotype.layer_count(), 2)

        duplicate_random_layer(genotype)
        self.assertEqual(genotype.layer_count(), 3)

        delete_random_layer(genotype)
        self.assertEqual(genotype.layer_count(), 2)

    def test_cell_mutation(self):
        genotype = Genotype()
        layer1 = GenotypeLayer()
        genotype.add_layer(layer1)
        self.assertEqual(genotype.layer_count(), 1)

        layer = genotype.get_layer(0)
        self.assertEqual(layer.cell_count(), 0)

        add_new_cell(genotype)
        layer = genotype.get_layer(0)
        self.assertEqual(layer.cell_count(), 1)
        add_new_cell(genotype)
        layer = genotype.get_layer(0)
        self.assertEqual(layer.cell_count(), 2)
