from typing import List

import torch
from torch import nn


class PhenotypeLayer(nn.Module):
    def __init__(self):
        super(PhenotypeLayer, self).__init__()
        self._cell_input_indices: List[
            (int, int)
        ] = []  # (left input index, right input index)
        self._cells = nn.ModuleList()

        self._memory: List[torch.Tensor] = []
        self._memory.append(torch.empty(0, 0))  # Just a dummy empty tensor

    def add_cell(self, left_input_index, right_input_index, cell):
        memory_len = len(self._memory)
        if left_input_index < memory_len and right_input_index < memory_len:
            self._cell_input_indices.append((left_input_index, right_input_index))
            self._cells.append(cell)
            self._memory.append(torch.empty(0, 0))  # Just a dummy empty tensor
        else:
            raise IndexError(
                "Attempt to access hidden state memory that hasn't been set yet"
            )

    # NOTE: The memory variable access is potentially dangerous. It might get accessed before there is data
    def forward(self, x):
        self._memory[0] = x

        for i in range(len(self._cell_input_indices)):
            cell_inputs = self._cell_input_indices[i]
            left = self._memory[cell_inputs[0]]
            right = self._memory[cell_inputs[1]]
            cell = self._cells[i]
            self._memory[i + 1] = cell(left, right)

        output = self._memory[-1]

        # Any unused hidden states are automatically added to the layer output via addition
        used_hidden_states = set()
        for (i, j) in self._cell_input_indices:
            used_hidden_states.add(i)
            used_hidden_states.add(j)
        all_hidden_states = set([i for i, _ in enumerate(self._memory[:-1])])

        unused_hidden_states = list(all_hidden_states - used_hidden_states)
        for state in unused_hidden_states:
            output += state

        return output
