import numpy as np
from matplotlib import pyplot as plt

# Noam optimizer is meant to extend Adam but ic can take in different kinds as well
# AdamW seems to work better than RAdam (possibly also better than bare Adam)
# AdamW combined with RMSELoss(reduction="sum") works really well


class NoamOptimizer:
    """
    Implements learning rate scheduler described in "Attention Is All You Need" paper
    lrate = embedding_size ** (−0.5) * min(step_num ** (−0.5), step_num * warmup_steps ** (−1.5))
    """

    def __init__(self, optimizer, embedding_dim, warmup_steps=4000):
        self.optimizer = optimizer
        self.embedding_dim = embedding_dim
        self.warmup_steps = warmup_steps

        self._current_step = 0
        self._current_rate = 0

    def step(self):
        self._current_step += 1

        rate = self.learning_rate()
        for parameter in self.optimizer.param_groups:
            parameter["lr"] = rate
        self._current_rate = rate

        self.optimizer.step()

    def learning_rate(self, step=None):
        if step is None:
            step = self._current_step

        return self.embedding_dim ** (-0.5) * min(
            step ** (-0.5),
            step * self.warmup_steps ** (-1.5),
        )

    def reset(self):
        self._current_step = 0
        self._current_rate = 0


if __name__ == "__main__":
    opts = [
        NoamOptimizer(None, 64, 500),
        NoamOptimizer(None, 64, 1000),
        NoamOptimizer(None, 32, 1000),
    ]
    plt.plot(
        np.arange(1, 10000),
        [[opt.learning_rate(i) for opt in opts] for i in range(1, 10000)],
    )
    plt.legend(["64:500", "64:1000", "32:1000"])
    plt.show()
