import copy
import unittest
from enum import Enum
from typing import List

import numpy as np

# MAE -> torch.nn.functional.l1_loss
# MSE -> torch.nn.functional.mse_loss
# RMSE -> sqrt(torch.nn.functional.mse_loss)


class Evaluation(str, Enum):
    MAE = "mae"
    MAPE = "mape"
    WAPE = "wape"
    MSE = "mse"
    RMSE = "rmse"

    @staticmethod
    def from_str(string: str):
        return Evaluation(string)

    @staticmethod
    def fn_list() -> List[Enum]:
        return [fn for fn in Evaluation]

    def fn(self):
        registry = {
            Evaluation.MAE: mae,
            Evaluation.MAPE: mape,
            Evaluation.WAPE: wape,
            Evaluation.MSE: mse,
            Evaluation.RMSE: rmse,
        }
        return copy.deepcopy(registry[self])


def mae(output, target):
    error = np.mean(np.abs(target - output))
    return error


def mape(output, target):
    diff = np.abs((target - output) / target)
    error = diff.mean()
    return error


# TODO: Double check the implementation, seems to produce weird numbers
def wape(output, target):
    diff = np.abs(target - output)
    error = diff.mean() / target.mean()
    return error


def mse(output, target):
    error = np.square(np.subtract(target, output)).mean()
    return error


def rmse(output, target):
    error = np.sqrt(np.square(np.subtract(target, output)).mean())
    return error


class TestEvaluation(unittest.TestCase):
    def test_wape(self):
        t1 = np.array([1.0, 2.0, 3.0])
        t2 = np.array([1.0, 2.0, 4.0])
        t3 = np.array([1.0, 20.0, 30.0])

        w1 = wape(t1, t1)
        w2 = wape(t1, t2)
        w3 = wape(t1, t3)

        self.assertAlmostEqual(w1, 0.0, 3)
        self.assertAlmostEqual(w2, 0.1428, 3)
        self.assertAlmostEqual(w3, 0.8823, 3)

    def test_mape(self):
        t1 = np.array([1.0, 2.0, 3.0])
        t2 = np.array([1.0, 2.0, 4.0])
        t3 = np.array([1.0, 20.0, 30.0])

        m1 = mape(t1, t1)
        m2 = mape(t1, t2)
        m3 = mape(t1, t3)

        self.assertAlmostEqual(m1, 0.0, 3)
        self.assertAlmostEqual(m2, 0.0833, 3)
        self.assertAlmostEqual(m3, 0.6, 3)

    def test_mse(self):
        t1 = np.array([1.0, 2.0, 3.0])
        t2 = np.array([1.0, 2.0, 4.0])
        t3 = np.array([1.0, 20.0, 30.0])

        m1 = mse(t1, t1)
        m2 = mse(t1, t2)
        m3 = mse(t1, t3)

        self.assertAlmostEqual(m1, 0.0, 3)
        self.assertAlmostEqual(m2, 0.333, 3)
        self.assertAlmostEqual(m3, 351.0, 3)

    def test_rmse(self):
        t1 = np.array([1.0, 2.0, 3.0])
        t2 = np.array([1.0, 2.0, 4.0])
        t3 = np.array([1.0, 20.0, 30.0])

        m1 = rmse(t1, t1)
        m2 = rmse(t1, t2)
        m3 = rmse(t1, t3)

        self.assertAlmostEqual(m1, 0.0, 3)
        self.assertAlmostEqual(m2, 0.577, 3)
        self.assertAlmostEqual(m3, 18.7349, 3)
