import copy
import torch.nn as nn

from src.components.attention.ScaledDotProductMultiHeadAttention import (
    ScaledDotProductMultiHeadAttention1,
)
from src.components.embedding.TimeSeriesPatchEmbedding import TimeSeriesPatchEmbedding
from src.components.transformer.Regression import Regression
from src.components.transformer.FeedForward import FeedForward
from src.transformers.decoder_only.StandaloneDecoder import StandaloneDecoderLayer
from src.transformers.decoder_only.StandaloneDecoder import StandaloneDecoder
from src.components.positional_encoding.ape.SinusoidalAbsolutePositionalEncoding import (
    SinusoidalAbsolutePositionalEncoding,
)


class TimeSeriesDecoderTransformer(nn.Module):
    def __init__(self, decoder, target_embedding, regression):
        super(TimeSeriesDecoderTransformer, self).__init__()
        self.decoder = decoder
        self.target_embedding = target_embedding
        self.regression = regression

    @staticmethod
    def create(
        decoder_layers=5,
        embedding_dim=20,
        forward_expansion=4,
        head_count=4,
        dropout=0.1,
        patch_size=1,
        xavier_param_init=True,
    ):
        attention = ScaledDotProductMultiHeadAttention1(embedding_dim, head_count)
        feed_forward = FeedForward(embedding_dim, forward_expansion, dropout)
        positional_encoding = SinusoidalAbsolutePositionalEncoding(
            embedding_dim, dropout
        )
        embedding = TimeSeriesPatchEmbedding(embedding_dim, patch_size)
        transformer = TimeSeriesDecoderTransformer(
            StandaloneDecoder(
                StandaloneDecoderLayer(
                    embedding_dim,
                    copy.deepcopy(attention),
                    copy.deepcopy(feed_forward),
                    dropout,
                ),
                decoder_layers,
            ),
            nn.Sequential(
                copy.deepcopy(embedding),
                copy.deepcopy(positional_encoding),
            ),
            Regression(embedding_dim, patch_size),
        )

        # Initialize parameters with Glorot / fan_avg
        if xavier_param_init:
            for p in transformer.parameters():
                if p.dim() > 1:
                    nn.init.xavier_uniform_(p)

        return transformer

    def get_regression(self):
        return self.regression

    def embedding_dim(self):
        return self.regression.embedding_dim

    def patch_size(self):
        return self.regression.patch_size

    def forward(self, target, target_mask):
        return self.decode(target, target_mask)

    def decode(self, target, target_mask):
        return self.decoder(self.target_embedding(target), target_mask)
