import copy
import time

import numpy as np
import torch
from colorama import Fore, Style
from matplotlib import pyplot as plt
from torch import nn

from src.training.TransformerLossCompute import TransformerLossCompute
from src.transformers.decoder_only.TimeSeriesDecoderTransformer import (
    TimeSeriesDecoderTransformer,
)
from src.dataset.SignalDataset import (
    SignalDataset,
    SignalType,
)
from src.training.NoamOptimizer import NoamOptimizer
from src.training.evaluation import mae, rmse
from src.utils.helpers import subsequent_mask


class Batch:
    def __init__(
        self, source, target, patch_size, padding_value=0.0, sequence_start_value=1.0
    ):
        concatenated = torch.cat([source, target], dim=1)
        extended = torch.cat(
            [
                torch.full(
                    (concatenated.size(0), patch_size, 1),
                    sequence_start_value,
                    device=concatenated.device,
                ),
                concatenated,
            ],
            dim=1,
        )
        self.target = extended[:, :-patch_size]
        self.y_target = extended[:, patch_size:]

        self.target_mask = self.make_target_mask(self.target, padding_value, patch_size)
        self.y_token_count = (self.y_target != padding_value).data.sum()

    @staticmethod
    def make_target_mask(target, padding, patch_size):
        """Create a mask to hide padding and future tokens"""
        target = target.squeeze(2)

        # NOTE: Padding mask is not supported for patch size >1
        if patch_size == 1:
            target_mask = (target != padding).unsqueeze(-2)
            subseq_mask = subsequent_mask(target.size(-1), target.device).type_as(
                target_mask.data
            )
            target_mask = target_mask & subseq_mask
        else:
            mask_size = target.size(-1) // patch_size
            target_mask = subsequent_mask(mask_size, target.device).type_as(target.data)
        return target_mask


def batch_iterator(dataset, batch_size, patch_size, device):
    for i in range(0, len(dataset) - batch_size, batch_size):
        batch = [dataset[j] for j in range(i, i + batch_size)]
        x = np.array([[[s] for s in b["x"]] for b in batch], dtype=np.dtype("f4"))
        y = np.array([[[s] for s in b["y"]] for b in batch], dtype=np.dtype("f4"))

        source = torch.tensor(x, requires_grad=True).to(device=device)
        target = torch.tensor(y, requires_grad=True).to(device=device)
        yield Batch(source, target, patch_size)


def run_epoch(dataset_iterator, model, loss_compute):
    """Helper for running one epoch"""
    total_tokens = 0
    total_loss = 0
    tokens = 0

    start = time.time()
    for i, batch in enumerate(dataset_iterator):
        output = model.forward(batch.target, batch.target_mask)
        loss = loss_compute(output, batch.y_target, batch.y_token_count)
        total_loss += loss
        tokens += batch.y_token_count
        total_tokens += batch.y_token_count

        if i % 50 == 1:
            elapsed = time.time() - start
            print(
                "> #%d, loss: %f, tokens per sec: %f"
                % (i, loss / batch.y_token_count, tokens / elapsed)
            )
            start = time.time()
            tokens = 0

    return total_loss / total_tokens


def make_prediction(model, source, predict_tokens, patch_size):
    source_y = copy.deepcopy(source)

    token_count = predict_tokens // patch_size
    for i in range(token_count):
        target = source_y

        mask_size = source_y.size(1) // patch_size
        target_mask = subsequent_mask(mask_size, source.device)

        output = model.decode(
            target,
            target_mask,
        )
        last_output = output[:, -1]
        value = model.regression(last_output)

        source_y = torch.cat([source_y, value.reshape((-1, patch_size))], dim=1)
    prediction = source_y[:, source.size(1) :]
    return prediction


def show_results(source, target, prediction):
    print("### Results ###")
    print(f"MAE: {mae(prediction, target)}")
    print(f"RMSE: {rmse(prediction, target)}")

    source_length = source.size
    target_length = target.size

    x_source = np.arange(0, source_length)
    x_target = np.arange(source_length, source_length + target_length)
    plt.plot(x_source, source)
    plt.plot(x_target, target)
    plt.plot(x_target, prediction)
    plt.show()


if __name__ == "__main__":
    print(Fore.LIGHTBLUE_EX, end="")
    print("### Train decoder only ###")

    device = "cuda" if torch.cuda.is_available() else "cpu"
    print(f"# CUDA: {torch.cuda.is_available()}")

    # Dataset
    x_length = 100
    y_length = 400
    dataset = SignalDataset(SignalType.SINE, x_length=x_length, y_length=y_length)
    print(f"# Dataset length: {len(dataset)}")
    train_set, test_set = torch.utils.data.random_split(dataset, [17000, 2500])

    # Model
    batch_size = 20
    embedding_dim = 16
    patch_size = 1
    criterion = nn.SmoothL1Loss(reduction="sum")
    model = TimeSeriesDecoderTransformer.create(
        decoder_layers=4,
        embedding_dim=embedding_dim,
        head_count=4,
        dropout=0.01,
        patch_size=patch_size,
    ).to(device)
    optimizer = torch.optim.AdamW(model.parameters(), lr=0, betas=(0.9, 0.98), eps=1e-9)
    scheduled_optimizer = NoamOptimizer(
        optimizer,
        embedding_dim,
        1000,
    )

    # Train
    epochs = 5
    for epoch in range(epochs):
        print(Fore.LIGHTRED_EX, end="")
        print(f"### Epoch: {epoch + 1} ###")

        start = time.time()

        print(Style.RESET_ALL, end="")
        model.train()
        train_loss = run_epoch(
            batch_iterator(train_set, batch_size, patch_size, device),
            model,
            TransformerLossCompute(model.regression, criterion, scheduled_optimizer),
        )
        print(Fore.LIGHTRED_EX, end="")
        print(f"## Train acc loss: {train_loss} ##")

        print(Fore.LIGHTYELLOW_EX, end="")
        model.eval()
        eval_loss = run_epoch(
            batch_iterator(test_set, batch_size, patch_size, device),
            model,
            TransformerLossCompute(model.regression, criterion, None),
        )
        print(Fore.LIGHTRED_EX, end="")
        print(f"## Eval acc loss: {eval_loss} ##")
        print(f"## Epoch took: {time.time() - start}s ##")
        print(Style.RESET_ALL, end="")

    model.eval()
    test_sample = test_set[0]
    target = test_sample["y"]
    source = test_sample["x"]
    source_tensor = torch.Tensor(source).unsqueeze(0).to(device=device)
    prediction_tensor = make_prediction(
        model, source_tensor, predict_tokens=y_length, patch_size=patch_size
    )
    prediction = prediction_tensor.squeeze(0).detach().cpu().numpy()
    show_results(source, target, prediction)
