import copy
import torch.nn as nn

from src.components.attention.ScaledDotProductMultiHeadAttention import (
    ScaledDotProductMultiHeadAttention1,
)
from src.components.embedding.TimeSeriesPatchEmbedding import TimeSeriesPatchEmbedding
from src.components.transformer.EncoderFirstElementRegression import (
    EncoderFirstElementRegression,
)
from src.components.transformer.EncoderRegression import EncoderRegression
from src.components.transformer.FeedForward import FeedForward
from src.transformers.encoder_only.StandaloneEncoder import StandaloneEncoderLayer
from src.transformers.encoder_only.StandaloneEncoder import StandaloneEncoder
from src.components.positional_encoding.ape.SinusoidalAbsolutePositionalEncoding import (
    SinusoidalAbsolutePositionalEncoding,
)


class TimeSeriesEncoderTransformer(nn.Module):
    def __init__(self, encoder, embedding, regression):
        super(TimeSeriesEncoderTransformer, self).__init__()
        self.encoder = encoder
        self.embedding = embedding
        self.regression = regression

    @staticmethod
    def create(
        layers=4,
        embedding_dim=20,
        input_dim=100,
        output_dim=100,
        forward_expansion=4,
        head_count=4,
        dropout=0.1,
        patch_size=1,
        use_full_regression=True,
        xavier_param_init=True,
    ):
        attention = ScaledDotProductMultiHeadAttention1(embedding_dim, head_count)
        feed_forward = FeedForward(embedding_dim, forward_expansion, dropout)
        positional_encoding = SinusoidalAbsolutePositionalEncoding(
            embedding_dim, dropout
        )
        embedding = TimeSeriesPatchEmbedding(embedding_dim, patch_size)
        if use_full_regression:
            regression = EncoderRegression(
                embedding_dim, input_dim, output_dim, patch_size
            )
        else:
            regression = EncoderFirstElementRegression(
                embedding_dim, input_dim, output_dim, patch_size
            )

        transformer = TimeSeriesEncoderTransformer(
            StandaloneEncoder(
                StandaloneEncoderLayer(
                    embedding_dim,
                    copy.deepcopy(attention),
                    copy.deepcopy(feed_forward),
                    dropout,
                ),
                layers,
            ),
            nn.Sequential(
                embedding,
                copy.deepcopy(positional_encoding),
            ),
            regression,
        )

        # Initialize parameters with Glorot / fan_avg
        if xavier_param_init:
            for p in transformer.parameters():
                if p.dim() > 1:
                    nn.init.xavier_uniform_(p)

        return transformer

    def get_regression(self):
        return self.regression

    def embedding_dim(self):
        return self.regression.embedding_dim

    def patch_size(self):
        return self.regression.patch_size

    def remove_positional_encoding(self):
        if isinstance(self.embedding, nn.Sequential):
            embedding = self.embedding[0]
            self.embedding = embedding

    def swap_attention(self, attention_module: nn.Module):
        self.encoder.swap_attention(attention_module)

    def forward(self, x, mask):
        return self.encode(x, mask)

    def encode(self, x, mask):
        return self.encoder(self.embedding(x), mask)
