import copy
import torch.nn as nn

from src.components.attention.ScaledDotProductMultiHeadAttention import (
    ScaledDotProductMultiHeadAttention,
)
from src.transformers.vanilla_unmodified.SimpleEmbedding import SimpleEmbedding
from src.components.transformer.FeedForward import FeedForward
from src.components.positional_encoding.ape.SinusoidalAbsolutePositionalEncoding import (
    SinusoidalAbsolutePositionalEncoding,
)
from src.components.transformer.Decoder import DecoderLayer, Decoder
from src.components.transformer.Encoder import EncoderLayer, Encoder
from src.transformers.vanilla_unmodified.Generator import Generator


class Transformer(nn.Module):
    def __init__(self, encoder, decoder, source_embedding, target_embedding, generator):
        super(Transformer, self).__init__()
        self.encoder = encoder
        self.decoder = decoder
        self.source_embedding = source_embedding
        self.target_embedding = target_embedding
        self.generator = generator

    @staticmethod
    def create_vanilla_transformer(
        source_vocabulary_size,
        target_vocabulary_size,
        encoder_layers=6,
        decoder_layers=6,
        embedding_dim=512,
        forward_expansion=4,
        head_count=8,
        dropout=0.1,
    ):
        attention = ScaledDotProductMultiHeadAttention(embedding_dim, head_count)
        feed_forward = FeedForward(embedding_dim, forward_expansion, dropout)
        positional_encoding = SinusoidalAbsolutePositionalEncoding(
            embedding_dim, dropout
        )
        transformer = Transformer(
            Encoder(
                EncoderLayer(
                    embedding_dim,
                    copy.deepcopy(attention),
                    copy.deepcopy(feed_forward),
                    dropout,
                ),
                encoder_layers,
            ),
            Decoder(
                DecoderLayer(
                    embedding_dim,
                    copy.deepcopy(attention),
                    copy.deepcopy(attention),
                    copy.deepcopy(feed_forward),
                    dropout,
                ),
                decoder_layers,
            ),
            nn.Sequential(
                SimpleEmbedding(embedding_dim, source_vocabulary_size),
                copy.deepcopy(positional_encoding),
            ),
            nn.Sequential(
                SimpleEmbedding(embedding_dim, target_vocabulary_size),
                copy.deepcopy(positional_encoding),
            ),
            Generator(embedding_dim, target_vocabulary_size),
        )

        # Initialize parameters with Glorot / fan_avg
        for p in transformer.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

        return transformer

    def forward(self, source, target, source_mask, target_mask):
        return self.decode(
            self.encode(source, source_mask), source_mask, target, target_mask
        )

    def encode(self, source, source_mask):
        return self.encoder(self.source_embedding(source), source_mask)

    def decode(self, memory, source_mask, target, target_mask):
        return self.decoder(
            self.target_embedding(target), memory, source_mask, target_mask
        )
