import json
from typing import List

import torch
from torch import nn
from torch.utils.tensorboard import SummaryWriter


# Start local tensorboard instance
# tensorboard --logdir=.runs

# If you want to log directly to S3 from local
# Set personal AWS profile: export AWS_PROFILE=personal
# Specify bucket as log_dir: tb = Tensorboard(log_dir="s3://master-thesis-2022/local")
# Start local tensorboard with data from a bucket: tensorboard --logdir=s3://master-thesis-2022/local

# SSH tunneling the Tensorflow interface from EC2. So it's possible to view it locally.
# ssh -i "transformer-instance-ssh-key.pem" -NL 6006:localhost:6006 ubuntu@<ec2-ip>


class Tensorboard:
    def __init__(self, log_dir: str = "./.runs/"):
        self.writer = SummaryWriter(log_dir=log_dir)

    def __del__(self):
        self.writer.close()

    def add_scalar(self, key, value, iteration):
        self.writer.add_scalar(key, value, iteration)

    def add_scalars(self, key, value_dict, iteration):
        self.writer.add_scalars(key, value_dict, iteration)

    def add_histogram(self, key, value_list, iteration):
        self.writer.add_histogram(key, value_list, iteration)

    def add_image(self, key, image, iteration):
        # image: (3, H, W)
        self.writer.add_image(key, image, iteration, dataformats="HWC")

    def add_images(self, key, image_list, iteration):
        # image_list: (N, 3, H, W)
        self.writer.add_images(key, image_list, iteration, dataformats="NHWC")

    def add_graph(self, model: nn.Module, model_input: List[torch.Tensor]):
        self.writer.add_graph(model, model_input)

    def add_text(self, key, text, iteration):
        self.writer.add_text(key, text, iteration)

    def add_json(self, key, data: dict, iteration, indent=4):
        jsn = json.dumps(data, indent=indent)
        text = "".join("\t" + line for line in jsn.splitlines(True))
        self.writer.add_text(key, text, iteration)

    def flush(self):
        self.writer.flush()
